#!/usr/bin/env python3
"""
Python script to re-run a single experiment locally.
"""
from argparse import ArgumentParser
from subprocess import call

from folktexts._io import load_json

from .experiments import Experiment


def setup_arg_parser() -> ArgumentParser:
    # Init parser
    parser = ArgumentParser(description="Re-run a single experiment from its JSON config file.")
    parser.add_argument(
        "--experiment-json",
        type=str,
        help="[string] Path to an experiment JSON file to load.",
        required=True,
    )
    # TODO: add over-writable key-word arguments

    return parser


if __name__ == '__main__':

    # Parse command-line arguments
    parser = setup_arg_parser()
    args = parser.parse_args()

    # Load experiment from JSON file
    print(f"Running experiment from config file at '{args.experiment_json}'...")
    exp = Experiment(**load_json(args.experiment_json))

    # Run the experiment
    cmdline_args = [
        cmd
        for key, value in exp.kwargs.items()
        for cmd in (f"--{key.replace('_', '-')}", str(value))
    ]

    call([exp.executable_path] + cmdline_args)



"""Common functions to use with transformer LLMs."""
from __future__ import annotations

import logging
import re
from pathlib import Path

import numpy as np
import torch
from transformers import AutoModelForCausalLM, AutoTokenizer


def query_model_batch(
    text_inputs: list[str],
    model: AutoModelForCausalLM,
    tokenizer: AutoTokenizer,
    context_size: int
) -> np.array:
    """Queries the model with a batch of text inputs.

    Parameters
    ----------
    text_inputs : list[str]
        The inputs to the model as a list of strings.
    model : AutoModelForCausalLM
        The model to query.
    tokenizer : AutoTokenizer
        The tokenizer used to encode the text inputs.
    context_size : int
        The maximum context size to consider for each input (in tokens).

    Returns
    -------
    np.array
        Model's last token probabilities for each input as a np.array of shape
        (len(text_inputs), vocab_size).
    """
    model_device = next(model.parameters()).device

    # Tokenize
    token_inputs = [tokenizer.encode(text, return_tensors="pt").flatten()[-context_size:] for text in text_inputs]
    idx_last_token = [tok_seq.shape[0] - 1 for tok_seq in token_inputs]

    # Pad
    tensor_inputs = torch.nn.utils.rnn.pad_sequence(
        token_inputs,
        batch_first=True,
        padding_value=tokenizer.pad_token_id,
    ).to(model_device)

    # Mask padded context
    attention_mask = tensor_inputs.ne(tokenizer.pad_token_id)

    # Query: run one forward pass, i.e., generate the next token
    with torch.no_grad():
        logits = model(input_ids=tensor_inputs, attention_mask=attention_mask).logits

    # Probabilities corresponding to the last token after the prompt
    last_token_logits = logits[torch.arange(len(idx_last_token)), idx_last_token]
    last_token_probs = torch.nn.functional.softmax(last_token_logits, dim=-1)
    return last_token_probs.to(dtype=torch.float16).cpu().numpy()


def add_pad_token(tokenizer):
    """Add a pad token to the model and tokenizer if it doesn't already exist.

    Here we're using the end-of-sentence token as the pad token. Both the model
    weights and tokenizer vocabulary are untouched.

    Another possible way would be to add a new token `[PAD]` to the tokenizer
    and update the tokenizer vocabulary and model weight embeddings accordingly.
    The embedding for the new pad token would be the average of all other
    embeddings.
    """
    if tokenizer.pad_token is None:
        tokenizer.add_special_tokens({'pad_token': tokenizer.eos_token})


def is_bf16_compatible() -> bool:
    """Checks if the current environment is bfloat16 compatible."""
    return torch.cuda.is_available() and torch.cuda.is_bf16_supported()


def load_model_tokenizer(model_name_or_path: str | Path, **kwargs) -> tuple[AutoModelForCausalLM, AutoTokenizer]:
    """Load a model and tokenizer from the given local path (or using the model name).

    Parameters
    ----------
    model_name_or_path : str | Path
        Model name or local path to the model folder.
    kwargs : dict
        Additional keyword arguments to pass to the model `from_pretrained` call.

    Returns
    -------
    tuple[AutoModelForCausalLM, AutoTokenizer]
        The loaded model and tokenizer, respectively.
    """
    logging.info(f"Loading model '{model_name_or_path}'")

    # Load tokenizer from disk
    tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)

    # Set default keyword arguments for loading the pretrained model
    model_kwargs = dict(
        torch_dtype=torch.bfloat16 if is_bf16_compatible() else torch.float16,
        trust_remote_code=True,
        device_map="auto",
    )
    model_kwargs.update(kwargs)

    # Load model from disk
    model = AutoModelForCausalLM.from_pretrained(
        model_name_or_path,
        **model_kwargs,
    )

    # Add pad token to the tokenizer if it doesn't already exist
    add_pad_token(tokenizer)

    # Move model to the correct device
    device = "cpu"
    if torch.cuda.is_available():
        device = "cuda"
    elif torch.backends.mps.is_available():
        device = "mps"
    logging.info(f"Moving model to device: {device}")
    if model.device.type != device:
        model.to(device)

    return model, tokenizer


def get_model_folder_path(model_name: str, root_dir="/tmp") -> str:
    """Returns the folder where the model is saved."""
    folder_name = model_name.replace("/", "--")
    return (Path(root_dir) / folder_name).resolve().as_posix()


def get_model_size_B(model_name: str, default: int = 2) -> int:
    """Get the model size from the model name, in Billions of parameters.
    """
    regex = re.search(r"((?P<times>\d+)[xX])?(?P<size>\d+)[bB]", model_name)
    if regex:
        return int(regex.group("size")) * int(regex.group("times") or 1)
    else:
        logging.error(f"Could not infer model size from name '{model_name}'")
    return default



"""Module to map risk-estimation questions to different prompting techniques.

e.g.,
- multiple-choice Q&A vs direct numeric Q&A;
- zero-shot vs few-shot vs CoT;
"""
import pandas as pd
from transformers import AutoTokenizer

from .dataset import Dataset
from .qa_interface import QAInterface
from .task import TaskMetadata

SYSTEM_PROMPT = """\
You are a helpful assistant. You answer multiple-choice questions based on the information provided.
"""

ACS_TASK_DESCRIPTION = """\
The following data corresponds to a survey respondent. \
The survey was conducted among US residents in 2018. \
Please answer the question based on the information provided. \
The data provided is enough to reach an approximate answer.
"""

ACS_FEW_SHOT_TASK_DESCRIPTION = """\
The following data corresponds to different survey respondents. \
The survey was conducted among US residents in 2018. \
Please answer each question based on the information provided. \
The data provided is enough to reach an approximate answer for each person.
"""

ANTHROPIC_CHAT_PROMPT = """If had to select one of the options, my answer would be"""
GEMMA_CHAT_PROMPT = """The provided information suggests that the answer is"""


def encode_row_prompt(
    row: pd.Series,
    task: TaskMetadata,
    add_task_description: bool = True,
    question: QAInterface = None,
) -> str:
    """Encode a question regarding a given row."""
    # Get the question to ask
    question = question or task.question
    return (
        (ACS_TASK_DESCRIPTION + "\n" if add_task_description else "")
        + f"""\
Information:
{task.get_row_description(row)}

{question.get_question_prompt()}""")


def encode_row_prompt_few_shot(
    row: pd.Series,
    task: TaskMetadata,
    dataset: Dataset,
    n_shots: int = 10,
    reuse_examples: bool = False,
    question: QAInterface = None,
) -> str:
    """Encode a question regarding a given row using few-shot prompting.

    Parameters
    ----------
    row : pd.Series
        The row that the question will be about.
    task : TaskMetadata
        The task that the row belongs to.
    n_shots : int, optional
        The number of example questions and answers to use before prompting
        about the given row, by default 3.
    reuse_examples : bool, optional
        Whether to reuse the same examples for consistency. By default will
        resample new examples each time (`reuse_examples=False`).

    Returns
    -------
    prompt : str
        The encoded few-shot prompt.
    """
    # Take `n_shots` random samples from the train set
    X_examples, y_examples = dataset.sample_n_train_examples(n_shots, reuse_examples=reuse_examples)

    # Start with task description
    prompt = ACS_FEW_SHOT_TASK_DESCRIPTION + "\n"

    # Get the question to ask
    question = question or task.question

    # Add `n` example rows with respective labels
    for i in range(n_shots):
        prompt += (
            encode_row_prompt(X_examples.iloc[i], task=task, add_task_description=False)
            + f" {question.get_answer_key_from_value(y_examples.iloc[i])}"
            + "\n\n"
        )

    # Add the target row without its label
    prompt += encode_row_prompt(
        row,
        task=task,
        add_task_description=False,
        question=question,
    )
    return prompt


def encode_row_prompt_chat(
    row: pd.Series,
    task: TaskMetadata,
    tokenizer: AutoTokenizer,
    question: QAInterface = None,
    **chat_template_kwargs,
) -> str:
    # TODO: implement two functions
    # - one for gemma-like models that are not compatible with system prompts
    # - and another for regular models compatible with system prompts
    return apply_chat_template(
        tokenizer,
        (
            SYSTEM_PROMPT
            + encode_row_prompt(row, task, question=question)
        ),
        **chat_template_kwargs,
    )


def apply_chat_template(
    tokenizer: AutoTokenizer,
    user_prompt: str,
    system_prompt: str = None,
    chat_prompt: str = ANTHROPIC_CHAT_PROMPT,
    **kwargs,
) -> str:
    # Add system prompt
    conversation = ([
        {"role": "system", "content": system_prompt}
    ] if system_prompt else [])

    # Add user prompt
    conversation.append({"role": "user", "content": user_prompt})

    # Using the Anthropic-style chat prompt
    conversation.append({"role": "assistant", "content": chat_prompt})

    # Default kwargs
    kwargs.setdefault("add_generation_prompt", False)

    # Apply prompt template
    filled_prompt = tokenizer.apply_chat_template(
        conversation=conversation,
        tokenize=False,
        **kwargs,
    )

    # Make sure no special tokens follow the `CHAT_PROMPT`;
    # > some models add a newline character and/or a <end_of_turn> token
    filled_prompt = filled_prompt[: len(chat_prompt) + filled_prompt.find(chat_prompt)]
    return filled_prompt



"""A collection of ACS prediction tasks based on the folktables package.
"""
from __future__ import annotations

from dataclasses import asdict, dataclass

import folktables
from folktables import BasicProblem

from .._utils import hash_dict
from ..col_to_text import ColumnToText as _ColumnToText
from ..task import TaskMetadata
from . import acs_columns

# Map of ACS column names to ColumnToText objects
_acs_columns_map: dict[str, object] = {
    col_mapper.name: col_mapper
    for col_mapper in acs_columns.__dict__.values()
    if isinstance(col_mapper, _ColumnToText)
}


@dataclass
class ACSTaskMetadata(TaskMetadata):
    """A class to hold information on an ACS prediction task."""

    # The ACS task object from the folktables package
    folktables_obj: BasicProblem = None

    @classmethod
    def make_folktables_task(
        cls,
        name: str,
        description: str,
        target_threshold: float | int = None,
    ) -> "ACSTaskMetadata":

        # Get the task object from the folktables package
        try:
            folktables_task = getattr(folktables, name)
        except AttributeError:
            raise ValueError(f"Could not find task '{name}' in folktables package.")

        acs_task = ACSTaskMetadata(
            name=name,
            description=description,
            features=folktables_task.features,
            target=folktables_task.target,
            cols_to_text=_acs_columns_map,
            sensitive_attribute=folktables_task.group,
            target_threshold=target_threshold,
            folktables_obj=folktables_task,
        )

        return acs_task

    def __hash__(self) -> int:
        hashable_params = asdict(self)
        hashable_params.pop("cols_to_text")
        hashable_params.pop("folktables_obj")
        return int(hash_dict(hashable_params), 16)


# Instantiate folktables tasks
acs_income_task = ACSTaskMetadata.make_folktables_task(
    name="ACSIncome",
    description="predict whether an individual's income is above $50,000",
    target_threshold=50000,
)

acs_public_coverage_task = ACSTaskMetadata.make_folktables_task(
    name="ACSPublicCoverage",
    description="predict whether an individual is covered by public health insurance",
)

acs_mobility_task = ACSTaskMetadata.make_folktables_task(
    name="ACSMobility",
    description="predict whether an individual had the same residential address one year ago",
)

acs_employment_task = ACSTaskMetadata.make_folktables_task(
    name="ACSEmployment",
    description="predict whether an individual is employed",
)

acs_travel_time_task = ACSTaskMetadata.make_folktables_task(
    name="ACSTravelTime",
    description="predict whether an individual has a commute to work that is longer than 20 minutes",
    target_threshold=20,
)



"""Module to plot evaluation results.
"""
from __future__ import annotations

import logging
from pathlib import Path
from typing import Callable

import numpy as np
import pandas as pd
from sklearn import metrics
from sklearn.calibration import CalibrationDisplay
from sklearn.metrics import RocCurveDisplay

from ._utils import safe_division

try:
    import matplotlib.pyplot as plt
    import seaborn as sns
    HAS_PLOTTING_DEPS = True
    sns.set_style("whitegrid", rc={"grid.linestyle": "--"})
except ImportError:
    HAS_PLOTTING_DEPS = False

_error_msg = (
    "Plotting functions will not work as optional plotting dependencies are not installed. "
    "Please install `matplotlib` and `seaborn` to enable plotting."
)


# Minimum fraction of the dataset size to consider a group for plotting
GROUP_SIZE_THRESHOLD = 0.03


def _check_plotting_deps() -> bool:
    if not HAS_PLOTTING_DEPS:
        logging.error(_error_msg)
    return HAS_PLOTTING_DEPS


def save_fig(fig, fig_name: str, imgs_dir: str | Path, format: str = "pdf") -> str:
    """Helper to save a matplotlib figure to disk."""
    path = Path(imgs_dir) / f"{fig_name}"
    path = path.with_suffix(f".{format}").resolve()
    fig.savefig(path, bbox_inches="tight", dpi=300)
    plt.close()
    return path.as_posix()


def render_evaluation_plots(
    y_true: np.ndarray,
    y_pred_scores: np.ndarray,
    *,
    eval_results: dict = {},
    model_name: str = None,
    imgs_dir: str | Path = None,
    show_plots: bool = False,
) -> dict:
    """Renders evaluation plots for the given predictions."""
    # Check if plotting dependencies are available
    if _check_plotting_deps() is False:
        return {}

    # Initialize vars
    results = {}
    model_str = f" - {model_name}" if model_name else ""

    # Helper function to show or save plot
    def show_or_save(fig, fig_name: str):
        if show_plots:
            plt.show()
        if imgs_dir:
            results[f"{fig_name}_path"] = save_fig(fig, fig_name, imgs_dir)

    # ### ### ### ###
    # Plot ROC curve
    # ### ### ### ###
    disp = RocCurveDisplay.from_predictions(y_true=y_true, y_pred=y_pred_scores, plot_chance_level=True)
    disp.figure_.suptitle("ROC Curve" + model_str)

    # Plot ROC point
    if "fpr" in eval_results and "tpr" in eval_results and "threshold" in eval_results:
        fpr, tpr = eval_results["fpr"], eval_results["tpr"]
        plt.plot(
            fpr, tpr, "ro", markersize=5, lw=0,
            label=f"threshold={eval_results['threshold']:.2f}")
        plt.legend()

    show_or_save(disp.figure_, "roc_curve")

    # ### ### ### ### ### ###
    # Plot calibration curve
    # ### ### ### ### ### ###
    disp = CalibrationDisplay.from_predictions(y_true=y_true, y_prob=y_pred_scores, n_bins=5, strategy="quantile")
    disp.figure_.suptitle("Calibration Curve" + model_str)
    show_or_save(disp.figure_, "calibration_curve")

    # ### ### ### ### ### ###
    # Plot score distribution
    # ### ### ### ### ### ###
    sns.histplot(y_pred_scores, bins=10, stat="frequency", kde=False)
    plt.xlabel("Predicted Risk Score")
    plt.ylabel("Frequency")
    plt.gcf().suptitle("Score Distribution" + model_str)
    show_or_save(plt.gcf(), "score_distribution")

    # ### ### ### ### ### ### ### ### ### ###
    # Plot distribution of scores per label #
    # ### ### ### ### ### ### ### ### ### ###
    sns.kdeplot(
        data=pd.DataFrame(
            {"score": y_pred_scores, "label": y_true}
        ).reset_index(drop=True),
        x="score",
        hue="label",
        multiple="fill",
    )
    plt.xlim(y_pred_scores.min(), y_pred_scores.max())
    plt.xlabel("Predicted Risk Score")
    plt.gcf().suptitle("Score Distribution per Label" + (f" - {model_name}" if model_name else ""))
    show_or_save(plt.gcf(), "score_distribution_per_label")

    return results


def render_fairness_plots(  # noqa: C901
    y_true: np.ndarray,
    y_pred_scores: np.ndarray,
    *,
    sensitive_attribute: np.ndarray,
    eval_results: dict = {},
    model_name: str = None,
    group_value_map: Callable[[int], str],
    group_size_threshold: float = GROUP_SIZE_THRESHOLD,
    imgs_dir: str | Path = None,
    show_plots: bool = False,
) -> dict:
    """Renders fairness plots for the given predictions."""
    # Check if plotting dependencies are available
    if _check_plotting_deps() is False:
        return {}

    # Plot fairness plots if sensitive attribute is provided
    assert len(sensitive_attribute) == len(y_true) == len(y_pred_scores), \
        "All arrays should have the same length."

    # Initialize vars
    results = {}
    model_str = f" - {model_name}" if model_name else ""
    n_groups = len(np.unique(sensitive_attribute))
    assert n_groups > 1, "At least 2 groups are required for fairness plots."

    # Helper function to show or save plot
    def show_or_save(fig, fig_name: str):
        if show_plots:
            plt.show()
        if imgs_dir:
            results[f"{fig_name}_path"] = save_fig(fig, fig_name, imgs_dir)

    # Set group-wise colors and global color
    palette = sns.color_palette(n_colors=n_groups + 1)
    global_color = palette[0]
    group_colors = palette[1:]
    group_line_styles = ["--", ":", "-."] * (1 + n_groups // 3)

    # ###
    # Plot group-specific ROC curves
    # ###
    for idx, s_value in enumerate(np.unique(sensitive_attribute)):
        is_first_group = (idx == 0)
        is_last_group = (idx == n_groups - 1)

        # If it's the first group
        if is_first_group:
            fig, ax = plt.subplots(figsize=(5, 4))

            # Plot global ROC point
            if "fpr" in eval_results and "tpr" in eval_results:
                ax.plot(
                    eval_results["fpr"], eval_results["tpr"],
                    marker="P", markersize=5, lw=0,
                    color=global_color, label="Global")

        # Get group-specific data
        group_indices = np.argwhere(sensitive_attribute == s_value).flatten()
        group_y_true = y_true[group_indices]
        group_y_pred_scores = y_pred_scores[group_indices]

        # If the group is too small of a fraction, skip (curve will be too erratic)
        if len(group_indices) / len(sensitive_attribute) < group_size_threshold:
            logging.info(f"Skipping group '{group_value_map(s_value)}' as it's too small.")
            continue

        # Plot group-specific ROC curve
        RocCurveDisplay.from_predictions(
            y_true=group_y_true, y_pred=group_y_pred_scores,
            plot_chance_level=is_last_group,
            ax=ax,

            # Group-specific visuals
            linestyle=group_line_styles[idx],
            color=group_colors[idx],
            name=group_value_map(s_value),
        )

        # Plot group-specific ROC point
        if "threshold" in eval_results:
            threshold = eval_results["threshold"]
            tn, fp, fn, tp = metrics.confusion_matrix(
                group_y_true, (group_y_pred_scores >= threshold).astype(int),
                labels=(0, 1),
            ).ravel()
            group_fpr = safe_division(fp, fp + tn, worst_result=1)
            group_tpr = safe_division(tp, tp + fn, worst_result=0)

            ax.plot(
                group_fpr, group_tpr,
                marker="X", markersize=5, lw=0,
                color=group_colors[idx],
                # label=group_value_map(s_value),
            )

    plt.legend()
    plt.title("ROC curve per sub-group" + model_str)
    show_or_save(fig, "roc_curve_per_subgroup")

    # ###
    # Plot group-specific calibration curves
    # ###
    for idx, s_value in enumerate(np.unique(sensitive_attribute)):
        group_indices = np.argwhere(sensitive_attribute == s_value).flatten()
        group_y_true = y_true[group_indices]
        group_y_pred_scores = y_pred_scores[group_indices]
        is_first_group = (idx == 0)

        if is_first_group:
            fig, ax = plt.subplots(figsize=(5, 4))

        # If the group is too small of a fraction, skip (curve will be too erratic)
        if len(group_indices) / len(sensitive_attribute) < group_size_threshold:
            logging.warning(f"Skipping group {s_value} plot as it's too small.")
            continue

        # Plot global calibration curve
        CalibrationDisplay.from_predictions(
            y_true=group_y_true, y_prob=group_y_pred_scores,
            n_bins=5, strategy="quantile",
            ax=ax,

            # Group-specific visuals
            linestyle=group_line_styles[idx],
            color=group_colors[idx],
            name=group_value_map(s_value),
        )

    plt.legend()
    plt.title("Calibration curve per sub-group" + model_str)
    show_or_save(fig, "calibration_curve_per_subgroup")

    # ###
    # Plot scores distribution per group
    # ###
    # TODO: make a decent score-distribution plot... # TODO: try score CDFs!
    # hist_bin_edges = np.histogram_bin_edges(y_pred_scores, bins=10)
    # for idx, s_value in enumerate(np.unique(sensitive_attribute)):
    #     group_indices = np.argwhere(sensitive_attribute == s_value).flatten()
    #     group_y_pred_scores = y_pred_scores[group_indices]
    #     is_first_group = (idx == 0)
    #     if is_first_group:
    #         fig, ax = plt.subplots()
    #     sns.histplot(
    #         group_y_pred_scores,
    #         bins=hist_bin_edges,
    #         stat="density",
    #         kde=False,
    #         color=group_colors[idx],
    #         label=group_value_map(s_value),
    #         ax=ax,
    #     )

    # plt.legend()
    # plt.title("Score distribution per sub-group" + model_str)
    # results["score_distribution_per_subgroup_path"] = save_fig(fig, "score_distribution_per_subgroup", imgs_dir)

    return results



"""Module to map risk-estimates to a variety of evaluation metrics.

Notes
-----
Code based on the `error_parity.evaluation` module,
at: https://github.com/socialfoundations/error-parity/blob/main/error_parity/evaluation.py
"""
from __future__ import annotations

import logging
import statistics
from typing import Optional

import numpy as np
from netcal.metrics import ECE
from sklearn.metrics import brier_score_loss, confusion_matrix, log_loss, roc_auc_score, roc_curve

from ._utils import is_valid_number, join_dictionaries, safe_division


def evaluate_binary_predictions(y_true: np.ndarray, y_pred: np.ndarray) -> dict:
    """Evaluates the provided binary predictions on common performance metrics.

    Parameters
    ----------
    y_true : np.ndarray
        The true class labels.
    y_pred : np.ndarray
        The binary predictions.

    Returns
    -------
    dict
        A dictionary with key-value pairs of (metric name, metric value).
    """
    # Compute confusion matrix
    tn, fp, fn, tp = confusion_matrix(y_true, y_pred, labels=(0, 1)).ravel()

    total = tn + fp + fn + tp
    pred_pos = tp + fp
    pred_neg = tn + fn
    assert pred_pos + pred_neg == total

    label_pos = tp + fn
    label_neg = tn + fp
    assert label_pos + label_neg == total

    results = {}

    # Accuracy
    results["accuracy"] = (tp + tn) / total

    # True Positive Rate (Recall)
    results["tpr"] = safe_division(tp, label_pos, worst_result=0)

    # False Negative Rate (1 - TPR)
    results["fnr"] = safe_division(fn, label_pos, worst_result=1)
    assert results["tpr"] + results["fnr"] == 1

    # False Positive Rate
    results["fpr"] = safe_division(fp, label_neg, worst_result=1)

    # True Negative Rate
    results["tnr"] = safe_division(tn, label_neg, worst_result=0)
    assert results["tnr"] + results["fpr"] == 1

    # Balanced accuracy
    results["balanced_accuracy"] = 0.5 * (results["tpr"] + results["tnr"])

    # Precision
    results["precision"] = safe_division(tp, pred_pos, worst_result=0)

    # Positive Prediction Rate
    results["ppr"] = safe_division(pred_pos, total, worst_result=0)

    return results


def evaluate_binary_predictions_fairness(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    sensitive_attribute: np.ndarray,
    return_groupwise_metrics: Optional[bool] = False,
) -> dict:
    """Evaluates fairness of the given predictions.

    Fairness metrics are computed as the ratios between group-wise performance
    metrics.

    Parameters
    ----------
    y_true : np.ndarray
        The true class labels.
    y_pred : np.ndarray
        The discretized predictions.
    sensitive_attribute : np.ndarray
        The sensitive attribute (protected group membership).
    return_groupwise_metrics : Optional[bool], optional
        Whether to return group-wise performance metrics (bool: True) or only
        the ratios between these metrics (bool: False), by default False.

    Returns
    -------
    dict
        A dictionary with key-value pairs of (metric name, metric value).
    """
    # All unique values for the sensitive attribute
    unique_groups = np.unique(sensitive_attribute)

    results = {}
    groupwise_metrics = {}
    unique_metrics = set()

    # Helper to compute key/name of a group-wise metric
    def group_metric_name(metric_name, group_name):
        return f"{metric_name}_group={group_name}"

    assert (
        len(unique_groups) > 1
    ), f"Found a single unique sensitive attribute: {unique_groups}"

    for s_value in unique_groups:
        # Indices of samples that belong to the current group
        group_indices = np.argwhere(sensitive_attribute == s_value).flatten()

        # Filter labels and predictions for samples of the current group
        group_labels = y_true[group_indices]
        group_preds = y_pred[group_indices]

        # Evaluate group-wise performance
        curr_group_metrics = evaluate_binary_predictions(group_labels, group_preds)

        # Add group-wise metrics to the dictionary
        groupwise_metrics.update(
            {
                group_metric_name(metric_name, s_value): metric_value
                for metric_name, metric_value in curr_group_metrics.items()
            }
        )

        unique_metrics = unique_metrics.union(curr_group_metrics.keys())

    # Compute ratios and absolute diffs
    for metric_name in unique_metrics:
        curr_metric_results = [
            groupwise_metrics[group_metric_name(metric_name, group_name)]
            for group_name in unique_groups
        ]

        # Metrics' ratio
        ratio_name = f"{metric_name}_ratio"

        # NOTE: should this ratio be computed w.r.t. global performance?
        # - i.e., min(curr_metric_results) / global_curr_metric_result;
        # - same question for the absolute diff calculations;
        results[ratio_name] = safe_division(
            min(curr_metric_results), max(curr_metric_results),
            worst_result=0,
        )

        # Metrics' absolute difference
        diff_name = f"{metric_name}_diff"
        results[diff_name] = max(curr_metric_results) - min(curr_metric_results)

    # ** Equalized odds **
    # default value: use maximum constraint violation for TPR and FPR equality
    results["equalized_odds_ratio"] = min(
        results["fnr_ratio"],
        results["fpr_ratio"],
    )
    results["equalized_odds_diff"] = max(
        results["tpr_diff"],  # same as FNR diff
        results["fpr_diff"],  # same as TNR diff
    )

    # Optionally, return group-wise metrics as well
    if return_groupwise_metrics:
        results.update(groupwise_metrics)

    return results


def compute_best_threshold(
    y_true: np.ndarray,
    y_pred_scores: np.ndarray,
    *,
    false_pos_cost: float = 1.0,
    false_neg_cost: float = 1.0,
) -> float:
    """Computes the binarization threshold that maximizes accuracy.

    Parameters
    ----------
    y_true : np.ndarray
        The true class labels.
    y_pred_scores : np.ndarray
        The predicted risk scores.
    false_pos_cost : float, optional
        The cost of a false positive error, by default 1.0
    false_neg_cost : float, optional
        The cost of a false negative error, by default 1.0

    Returns
    -------
    best_threshold : float
        The threshold value that maximizes accuracy for the given predictions.
    """

    # Compute TPR and FPR for all possible thresholds
    fpr, tpr, thresholds = roc_curve(y_true, y_pred_scores)

    # Compute the cost of each threshold
    costs = false_pos_cost * fpr + false_neg_cost * (1 - tpr)

    # Get the threshold that minimizes the cost
    best_threshold = thresholds[np.argmin(costs)]
    return best_threshold


def evaluate_predictions(
    y_true: np.ndarray,
    y_pred_scores: np.ndarray,
    *,
    sensitive_attribute: np.ndarray = None,
    threshold: float | str = "best",
    model_name: str = None,
) -> dict:
    """Evaluates predictions on common performance and fairness metrics.

    Parameters
    ----------
    y_true : np.ndarray
        The true class labels.
    y_pred_scores : np.ndarray
        The predicted scores.
    sensitive_attribute: np.ndarray, optional
        The sensitive attribute data. Will compute fairness metrics if provided.
    threshold : float | str, optional
        The threshold to use for binarizing the predictions, or "best" to infer
        which threshold maximizes accuracy.
    model_name : str, optional
        The name of the model to be used on the plots, by default None.

    Returns
    -------
    results : dict
        A dictionary with key-value pairs of (metric name, metric value).
    """
    # Compute threshold if necessary
    if threshold == "best":
        threshold = compute_best_threshold(y_true, y_pred_scores)
    assert is_valid_number(threshold) and 0 <= threshold <= 1, \
        f"Invalid threshold: {threshold}"

    # Save initial results' statistics
    results = {
        "threshold": threshold,
        "n_samples": len(y_true),
        "n_positives": np.sum(y_true).item(),
        "n_negatives": len(y_true) - np.sum(y_true).item(),
        "model_name": model_name,
    }

    # Compute binary predictions using the default threshold
    y_pred_binary = (y_pred_scores >= threshold).astype(int)

    # Evaluate binary predictions
    results.update(evaluate_binary_predictions(y_true, y_pred_binary))

    # Add loss functions as proxies for calibration
    results["log_loss"] = log_loss(y_true, y_pred_scores, labels=[0, 1])
    results["brier_score_loss"] = brier_score_loss(y_true, y_pred_scores)

    # Evaluate fairness metrics
    if sensitive_attribute is not None:
        results.update(evaluate_binary_predictions_fairness(y_true, y_pred_binary, sensitive_attribute))

    # Compute additional metrics
    results["roc_auc"] = roc_auc_score(y_true, y_pred_scores)

    # Compute Expected Calibration Error
    # TODO: re-implement ECE scorer to avoid including 10 other dependencies for this one metric...
    class_preds = np.stack([1 - y_pred_scores, y_pred_scores], axis=1)
    n_bins = 10
    results["ece"] = ECE(bins=n_bins, equal_intervals=True).measure(class_preds, y_true)

    # Try to compute ECE with quantile bins
    try:
        results["ece_quantile"] = ECE(bins=n_bins, equal_intervals=False).measure(class_preds, y_true)
    except Exception as err:
        logging.warning(f"Failed to compute ECE quantile: {err}")
        results["ece_quantile"] = None

    return results


def evaluate_predictions_bootstrap(
    y_true: np.ndarray,
    y_pred_scores: np.ndarray,
    *,
    sensitive_attribute: np.ndarray = None,
    threshold: float | str = "best",
    k: int = 200,
    confidence_pct: float = 95,
    seed: int = 42,
) -> dict:
    """Computes bootstrap estimates of several metrics for the given predictions.

    Parameters
    ----------
    y_true : np.ndarray
        The true labels.
    y_pred_scores : np.ndarray
        The score predictions.
    sensitive_attribute : np.ndarray, optional
        The sensitive attribute data. Will compute fairness metrics if provided.
    threshold : float | str, optional
        The threshold to use for binarizing the predictions, or "best" to infer
        which threshold maximizes accuracy, by default "best".
    k : int, optional
        How many bootstrap samples to draw, by default 200.
    confidence_pct : float, optional
        How large of a confidence interval to use when reporting lower and upper
        bounds, by default 95 (i.e., 2.5 to 97.5 percentile of results).
    seed : int, optional
        The random seed, by default 42.

    Returns
    -------
    results : dict
        A dictionary containing bootstrap estimates for a variety of metrics.
    """
    assert len(y_true) == len(y_pred_scores)
    rng = np.random.default_rng(seed=seed)

    # Draw k bootstrap samples with replacement
    results = []
    for _ in range(k):
        # Indices of current bootstrap sample
        indices = rng.choice(len(y_true), replace=True, size=len(y_true))

        # Evaluate predictions on this bootstrap sample
        results.append(
            evaluate_predictions(
                y_true=y_true[indices],
                y_pred_scores=y_pred_scores[indices],
                sensitive_attribute=sensitive_attribute[indices] if sensitive_attribute is not None else None,
                threshold=threshold,
            )
        )

    # Compute statistics from bootstrapped results
    all_metrics = set(results[0].keys())

    bt_mean = {}
    bt_stdev = {}
    bt_percentiles = {}

    low_percentile = (100 - confidence_pct) / 2
    confidence_percentiles = [low_percentile, 100 - low_percentile]

    for m in all_metrics:
        metric_values = [r[m] for r in results]

        bt_mean[m] = statistics.mean(metric_values)
        bt_stdev[m] = statistics.stdev(metric_values)
        bt_percentiles[m] = tuple(np.percentile(metric_values, confidence_percentiles))

    # Join bootstrap results for all metrics
    return join_dictionaries(
        *(
            {
                f"{metric}_mean": bt_mean[metric],
                f"{metric}_stdev": bt_stdev[metric],
                f"{metric}_low-percentile": bt_percentiles[metric][0],
                f"{metric}_high-percentile": bt_percentiles[metric][1],
            }
            for metric in sorted(bt_mean.keys())
        )
    )



#!/usr/bin/env python3
import gc
import logging
from argparse import ArgumentParser
from pathlib import Path

import torch
from tqdm import tqdm

# Temporary directory to use as download cache
TMP_DIR = Path("/tmp/")

# Default list of models to download
DEFAULT_MODEL_LIST = [
    # ** Small models **
    "openai-community/gpt2",
    "openai-community/gpt2-large",
    "openai-community/gpt2-xl",
    "google/gemma-2b",
    "google/gemma-1.1-2b-it",

    # ** Medium models **
    "google/gemma-7b",
    "google/gemma-1.1-7b-it",
    "mistralai/Mistral-7B-v0.1",
    "mistralai/Mistral-7B-Instruct-v0.2",
    "meta-llama/Meta-Llama-3-8B",
    "meta-llama/Meta-Llama-3-8B-Instruct",

    # ** Large models **
    "01-ai/Yi-34B",
    "01-ai/Yi-34B-Chat",
    "mistralai/Mixtral-8x7B-v0.1",
    "mistralai/Mixtral-8x7B-Instruct-v0.1",
    "meta-llama/Meta-Llama-3-70B",
    "meta-llama/Meta-Llama-3-70B-Instruct",
    "Qwen/Qwen1.5-72B",
    "Qwen/Qwen1.5-72B-Chat",
    "allenai/tulu-2-dpo-70b",
    "mistralai/Mixtral-8x22B-v0.1",
    "mistralai/Mixtral-8x22B-Instruct-v0.1",
]


def setup_arg_parser() -> ArgumentParser:
    # Init parser
    parser = ArgumentParser(description="Download huggingface transformer models and tokenizers to disk")

    parser.add_argument(
        "--model", type=str,
        help="[string] Model name on huggingface hub - can provide multiple models!",
        required=False,
        action="append",
    )

    parser.add_argument(
        "--save-dir", type=str,
        help="[string] Directory to save the downloaded models to",
        required=True,
    )

    parser.add_argument(
        "--tmp-cache-dir", type=str,
        help="[string] Cache dir to temporarily download models to",
        required=False,
        default=TMP_DIR,
    )

    return parser


def is_bf16_compatible() -> bool:
    """Checks if the current environment is bfloat16 compatible."""
    return torch.cuda.is_available() and torch.cuda.is_bf16_supported()


if __name__ == "__main__":
    # Parse command-line arguments
    parser = setup_arg_parser()
    args = parser.parse_args()

    # Process arguments
    model_list = args.model or DEFAULT_MODEL_LIST
    save_dir = Path(args.save_dir).expanduser().resolve()
    save_dir.mkdir(exist_ok=True, parents=False)

    cache_dir = Path(args.tmp_cache_dir).expanduser().resolve()
    cache_dir.mkdir(exist_ok=True, parents=False)

    for model_name in tqdm(model_list):
        # Create sub-folder to save this model to
        from folktexts.llm_utils import get_model_folder_path
        curr_save_dir = get_model_folder_path(model_name, root_dir=save_dir)

        # If model already exists on disk, skip
        if Path(curr_save_dir).exists():
            logging.warning(f"Model '{model_name}' already exists at '{curr_save_dir}'")
            continue

        # Download model to tmp dir
        from folktexts.llm_utils import load_model_tokenizer
        model, tokenizer = load_model_tokenizer(model_name, cache_dir=cache_dir)

        # Save model and tokenizer to disk
        print(f"Saving {model_name} to {curr_save_dir}")
        model.save_pretrained(curr_save_dir)
        tokenizer.save_pretrained(curr_save_dir)

        # Delete references to the model and tokenizer and force garbage collection
        del model
        del tokenizer
        gc.collect()

        # Empty VRAM if GPU is available
        if torch.cuda.is_available():
            torch.cuda.empty_cache()



"""Common set of utility functions and constants used across the project."""
from __future__ import annotations

import hashlib
import json
import logging
import operator
from datetime import datetime
from functools import partial, reduce
from pathlib import Path

import numpy as np


def is_valid_number(num) -> bool:
    """Check if the given number is a valid numerical value and not NaN."""
    return isinstance(num, (float, int, np.number)) and not np.isnan(num)


def safe_division(a: float, b: float, *, worst_result: float):
    """Try to divide the given arguments and return `worst_result` if unsuccessful."""
    if b == 0 or not is_valid_number(a) or not is_valid_number(b):
        logging.debug(
            f"Using `worst_result={worst_result}` in place of the following "
            f"division: {a} / {b}")
        return worst_result
    else:
        return a / b


def join_dictionaries(*dicts) -> dict:
    """Joins a sequence of dictionaries into a single dictionary."""
    return reduce(operator.or_, dicts)


def get_current_timestamp() -> str:
    """Return a timestamp representing the current time up to the second."""
    return datetime.now().strftime("%Y.%m.%d-%H.%M.%S")


def get_current_date() -> str:
    """Return a timestamp representing the current time up to the second."""
    return datetime.now().strftime("%Y-%m-%d")


def hash_dict(d: dict, length: int = 8) -> str:
    """Hashes a dictionary using SHAKE-256 and returns the hexdigest.

    Parameters
    ----------
    d : dict
        The dictionary to hash.
    length : int, optional
        The length of the hexdigest in number of text characters, by default 8.

    Returns
    -------
    hexdigest : str
        A string representing the hash in hexadecimal format.
    """
    d_enc = json.dumps(d, sort_keys=True).encode()
    return hashlib.shake_256(d_enc).hexdigest(length // 2)


def hash_function(func, length: int = 8) -> str:
    """Hashes a function using SHAKE-256 and returns the hexdigest.

    Parameters
    ----------
    func : Callable
        The function to hash. Can be a partial function.
    length : int, optional
        The length of the hexdigest in number of text characters, by default 8.

    Returns
    -------
    hexdigest : str
        A string representing the hash in hexadecimal format.
    """
    if not callable(func):
        raise ValueError("The input must be a callable function.")
    elif isinstance(func, partial):
        return hash_function(func.func)

    func_str = func.__code__.co_code
    return hashlib.shake_256(func_str).hexdigest(length // 2)


def standardize_path(path: str | Path) -> str:
    """Represents a posix path as a standardized string."""
    return Path(path).expanduser().resolve().as_posix()



"""Utils for the folktexts cmd-line interface.
"""
from __future__ import annotations


def cmd_line_args_to_kwargs(cmdline_args: list) -> dict:
    """Converts a list of command-line arguments to a dictionary of keyword arguments."""
    def _handle_str_value(val: str) -> int | float | str | bool:
        # Try bool
        if val.lower() in ("true", "false"):
            return val.lower() == "true"

        # Try int
        try:
            return int(val)
        except ValueError:
            pass

        # Try float
        try:
            return float(val)
        except ValueError:
            pass

        # Otherwise, assume it's a string
        return val

    kwargs_dict = {}
    for arg in cmdline_args:
        parsed_arg = arg.lstrip("-").replace("-", "_")
        if "=" in parsed_arg:
            split_idx = parsed_arg.index("=")
            key = parsed_arg[:split_idx]
            val = parsed_arg[split_idx + 1:]
            kwargs_dict[key] = _handle_str_value(val)
        else:
            kwargs_dict[parsed_arg] = True

    return kwargs_dict



from __future__ import annotations

import logging
import math
from functools import partial
from pathlib import Path
from typing import Callable

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, ClassifierMixin
from tqdm.auto import tqdm
from transformers import AutoModelForCausalLM, AutoTokenizer

from folktexts.dataset import Dataset
from folktexts.evaluation import compute_best_threshold
from folktexts.llm_utils import query_model_batch
from folktexts.prompting import encode_row_prompt as default_encode_row_prompt
from folktexts.qa_interface import DirectNumericQA, MultipleChoiceQA
from folktexts.task import TaskMetadata

from ._utils import hash_dict, hash_function

DEFAULT_CONTEXT_SIZE = 500
DEFAULT_BATCH_SIZE = 16
DEFAULT_CORRECT_ORDER_BIAS = True

SCORE_COL_NAME = "risk_score"
LABEL_COL_NAME = "label"


class LLMClassifier(BaseEstimator, ClassifierMixin):
    """An interface to use a transformer-based LLM as a binary classifier."""

    def __init__(
        self,
        model: AutoModelForCausalLM,
        tokenizer: AutoTokenizer,
        task: TaskMetadata | str,
        encode_row: Callable[[pd.Series], str] = None,
        correct_order_bias: bool = DEFAULT_CORRECT_ORDER_BIAS,
        threshold: float = 0.5,
        **inference_kwargs,
    ):
        """Creates an LLMClassifier object.

        Parameters
        ----------
        model : AutoModelForCausalLM
            The torch language model to use for inference.
        tokenizer : AutoTokenizer
            The tokenizer used to train the model.
        task : TaskMetadata | str
            The task metadata object or name of an already created task. This
            will be used to encode tabular rows into natural text prompts.
        encode_row : Callable[[pd.Series], str], optional
            The function used to encode tabular rows into natural text. If not
            provided, will use the default encoding function for the task.
        threshold : float, optional
            The classification threshold to use when outputting binary
            predictions, by default 0.5. Must be between 0 and 1. Will be
            re-calibrated if `fit` is called.
        **inference_kwargs
            Additional keyword arguments to pass to `self.predict_proba(...)`
            during inference. By default, will use `context_size=500` and
            `batch_size=16`.
        """
        self._model = model
        self._tokenizer = tokenizer
        self._task = TaskMetadata.get_task(task) if isinstance(task, str) else task
        self._encode_row = encode_row or partial(
            default_encode_row_prompt,
            task=self.task,
        )
        self._threshold = threshold

        self.correct_order_bias = correct_order_bias

        # Default inference kwargs
        self._inference_kwargs = inference_kwargs
        self._inference_kwargs.setdefault("context_size", DEFAULT_CONTEXT_SIZE)
        self._inference_kwargs.setdefault("batch_size", DEFAULT_BATCH_SIZE)

    @property
    def model(self) -> AutoModelForCausalLM:
        return self._model

    @property
    def tokenizer(self) -> AutoTokenizer:
        return self._tokenizer

    @property
    def task(self) -> TaskMetadata:
        return self._task

    @property
    def encode_row(self) -> Callable[[pd.Series], str]:
        return self._encode_row

    @property
    def threshold(self) -> float:
        return self._threshold

    @property
    def model_name(self) -> str:
        return Path(self.model.name_or_path).name

    @threshold.setter
    def threshold(self, value: float) -> float:
        if not 0 <= value <= 1:
            logging.error(f"Threshold must be between 0 and 1; got {value}.")

        # Clip threshold to valid range
        self._threshold = np.clip(value, 0, 1)
        logging.info(f"Set threshold to {self._threshold}.")

    def __hash__(self) -> int:
        """Generate a unique hash for the LLMClassifier object."""

        # All parameters that affect the model's behavior
        hash_params = dict(
            model_name=self.model_name,
            model_size=self.model.num_parameters(),
            tokenizer_vocab_size=self.tokenizer.vocab_size,
            task_hash=hash(self.task),
            correct_order_bias=self.correct_order_bias,
            threshold=self.threshold,
            encode_row_hash=hash_function(self.encode_row),
        )

        return int(hash_dict(hash_params), 16)

    def fit(self, X, y, *, false_pos_cost=1.0, false_neg_cost=1.0):
        """Uses the provided data sample to fit the prediction threshold."""

        # Compute risk estimates for the data
        y_pred_scores = self.predict_proba(X)
        if len(y_pred_scores.shape) > 1:
            y_pred_scores = y_pred_scores[:, -1]

        # Compute the best threshold for the given data
        self.threshold = compute_best_threshold(
            y, y_pred_scores,
            false_pos_cost=false_pos_cost,
            false_neg_cost=false_neg_cost,
        )

        # Update sklearn is_fitted status
        self._is_fitted = True
        return self

    def __sklearn_is_fitted__(self):
        """Check fitted status and return a Boolean value."""
        return hasattr(self, "_is_fitted") and self._is_fitted

    @staticmethod
    def _get_positive_class_scores(risk_scores: np.ndarray) -> np.ndarray:
        """Helper function to get positive class scores from risk scores."""
        if len(risk_scores.shape) > 1:
            return risk_scores[:, -1]
        else:
            return risk_scores

    def predict(
        self,
        data: pd.DataFrame | Dataset,
        batch_size: int = None,
        context_size: int = None,
        predictions_save_path: str | Path = None,
        labels: pd.Series | np.ndarray = None,
    ) -> np.ndarray | dict[str, np.ndarray]:
        """Returns binary predictions for the given data."""
        risk_scores = self.predict_proba(
            data,
            batch_size=batch_size,
            context_size=context_size,
            predictions_save_path=predictions_save_path,
            labels=labels,
        )
        if isinstance(risk_scores, dict):
            return {
                data_type: (self._get_positive_class_scores(data_scores) >= self.threshold).astype(int)
                for data_type, data_scores in risk_scores.items()
            }
        else:
            return (self._get_positive_class_scores(risk_scores) >= self.threshold).astype(int)

    def _load_predictions_from_disk(
        self,
        predictions_save_path: str | Path,
        data: pd.DataFrame | Dataset,
    ) -> np.ndarray | dict[str, np.ndarray] | None:
        """Attempts to load pre-computed predictions from disk."""
        predictions_save_path = Path(predictions_save_path)

        # If DF, try to load predictions as a CSV file
        if isinstance(data, pd.DataFrame):
            predictions_save_path = predictions_save_path.with_suffix(".csv")
            predictions_df = pd.read_csv(predictions_save_path, index_col=0)

            # Check if index matches our current dataframe
            if predictions_df.index.equals(data.index):
                return predictions_df[SCORE_COL_NAME].values
            else:
                logging.error("Saved predictions do not match the current dataframe.")

        # If Dataset, try to load predictions as a pickled dict
        elif isinstance(data, Dataset):
            from ._io import load_pickle
            predictions_save_path = predictions_save_path.with_suffix(".pkl")
            predictions_dict = load_pickle(predictions_save_path)

            # Check if the predictions' indices match the current dataset
            if all(
                preds_array.index.equals(data.get_data_split(data_type)[0].index)
                for data_type, preds_array in predictions_dict.items()
            ):
                return {
                    data_type: predictions_dict[data_type][SCORE_COL_NAME].values
                    for data_type in predictions_dict.keys()
                }
            else:
                logging.error("Saved predictions do not match the current dataset splits.")

        else:
            logging.error(f"Cannot load predictions from disk for data type {type(data)}.")
            return None

    def predict_proba(
        self,
        data: pd.DataFrame | Dataset,
        batch_size: int = None,
        context_size: int = None,
        predictions_save_path: str | Path = None,
        labels: pd.Series | np.ndarray = None,
    ) -> np.ndarray | dict[str, np.ndarray]:
        """Returns probability estimates for the given data.

        Parameters
        ----------
        data : pd.DataFrame | Dataset
            The data to compute risk estimates for. Can be a pandas DataFrame or
            a Dataset object. If a Dataset object is provided, will compute risk
            scores for all available data splits.
        batch_size : int, optional
            The batch size to use when running inference.
        context_size : int, optional
            The context size to use when running inference (in number of tokens).
        predictions_save_path : str | Path, optional
            If provided, will save the computed risk scores to this path in
            disk. If the path exists, will attempt to load pre-computed
            predictions from this path.
        labels : pd.Series | np.ndarray, optional
            The labels corresponding to the provided data. Not required to
            compute predictions. Will only be used to save alongside predictions
            to disk.

        Returns
        -------
        risk_scores : np.ndarray | dict[str, np.ndarray]
            The risk scores for the given data, or a dictionary of data split
            name to risk scores if a Dataset object was provided.
        """
        # Validate arguments
        if labels is not None and predictions_save_path is None:
            logging.error(
                "** Ignoring `labels` argument as `predictions_save_path` was not provided. **"
                "The `labels` argument is only used in conjunction with "
                "`predictions_save_path` to save alongside predictions to disk. ")

        # Check if `predictions_save_path` exists and load predictions if possible
        if predictions_save_path is not None and Path(predictions_save_path).exists():
            result = self._load_predictions_from_disk(predictions_save_path, data=data)
            if result is not None:
                logging.info(f"Loaded predictions from {predictions_save_path}.")
                return result
            else:
                logging.error(
                    f"Failed to load predictions from {predictions_save_path}. "
                    f"Re-computing predictions and overwriting local file..."
                )

        # Compute risk estimates
        # (if local save path was not provided or does not match current data)
        if isinstance(data, pd.DataFrame):
            risk_scores = self._compute_risk_estimates_for_dataframe(
                df=data,
                batch_size=batch_size,
                context_size=context_size,
            )

            # Save to disk if `predictions_save_path` is provided
            if predictions_save_path is not None:
                predictions_save_path = Path(predictions_save_path).with_suffix(".csv")

                predictions_df = pd.DataFrame(risk_scores, index=data.index, columns=[SCORE_COL_NAME])
                predictions_df[LABEL_COL_NAME] = labels
                predictions_df.to_csv(predictions_save_path, index=True, mode="w")

            return risk_scores

        elif isinstance(data, Dataset):
            # TODO: save predictions in a standardized file format when given a Dataset
            # > we have the dataset name, splits, and seed, so we can safely save predictions for future use
            scores_dict = self._compute_risk_estimates_for_dataset(
                dataset=data,
                batch_size=batch_size,
                context_size=context_size,
            )

            # Save to disk if `predictions_save_path` is provided
            if predictions_save_path is not None:
                predictions_save_path = Path(predictions_save_path).with_suffix(".pkl")

                from ._io import save_pickle
                logging.warning(    # TODO
                    f"Saving dataset predictions to {predictions_save_path.as_posix()}. "
                    f"TODO: remove pickling functionality and save everything as csv files "
                    f"to re-use file-handling code from `_compute_risk_estimates_for_dataframe`."
                )
                save_pickle(scores_dict, predictions_save_path)

            return scores_dict

        else:
            raise ValueError(
                f"`data` must be a pandas DataFrame or a Dataset object; "
                f"received {type(data)} instead."
            )

    def _compute_risk_estimates_for_dataframe(
        self,
        df: pd.DataFrame,
        batch_size: int = None,
        context_size: int = None,
    ) -> np.ndarray:
        """Compute risk estimates for a specific dataframe (internal helper function).

        Parameters
        ----------
        df : pd.DataFrame
            The dataframe to compute risk estimates for.
        batch_size : int, optional
            The batch size to use when running inference.
        context_size : int, optional
            The context size to use when running inference (in number of tokens).

        Returns
        -------
        risk_scores : np.ndarray
            The risk estimates for each row in the dataframe.
        """

        # Initialize risk scores and other constants
        fill_value = -1
        risk_scores = np.empty(len(df))
        risk_scores.fill(fill_value)    # fill with -1's

        batch_size = batch_size or self._inference_kwargs["batch_size"]
        context_size = context_size or self._inference_kwargs["context_size"]
        num_batches = math.ceil(len(df) / batch_size)

        # Get questions to ask
        q = self.task.question
        questions = [q]
        if self.correct_order_bias:
            if isinstance(q, DirectNumericQA):
                logging.info("No need to correct ordering bias for DirectNumericQA prompting.")
            elif isinstance(q, MultipleChoiceQA):
                questions = list(MultipleChoiceQA.create_answer_keys_permutations(q))
            else:
                logging.error(f"Unknown question type '{type(q)}'; cannot correct ordering bias.")

        # Compute risk estimates per batch
        for batch_idx in tqdm(range(num_batches), desc="Computing risk estimates"):
            start_idx = batch_idx * batch_size
            end_idx = min((batch_idx + 1) * batch_size, len(df))
            batch_data = df.iloc[start_idx:end_idx]

            batch_risk_scores = np.empty((len(batch_data), len(questions)))
            for idx, q in enumerate(questions):

                # Encode batch data into natural text prompts
                data_texts_batch = [
                    self.encode_row(row, question=q)
                    for _, row in batch_data.iterrows()
                ]

                # Query model
                last_token_probs_batch = query_model_batch(
                    data_texts_batch,
                    self.model,
                    self.tokenizer,
                    context_size=context_size,
                )

                # Decode model output
                risk_estimates_batch = [
                    q.get_answer_from_model_output(
                        ltp,
                        tokenizer=self.tokenizer,
                    )
                    for ltp in last_token_probs_batch
                ]

                # Store risk estimates for current question
                batch_risk_scores[:, idx] = risk_estimates_batch

            risk_scores[start_idx:end_idx] = batch_risk_scores.mean(axis=1)

        # Check that all risk scores were computed
        assert not np.isclose(risk_scores, fill_value).any()
        return risk_scores

    def _compute_risk_estimates_for_dataset(
        self,
        dataset: Dataset,
        batch_size: int = None,
        context_size: int = None,
    ):
        """Computes risk estimates for each row in the dataset.

        Parameters
        ----------
        dataset : Dataset
            The dataset to compute risk estimates for.
        batch_size : int, optional
            The batch size to use when running inference.
        context_size : int, optional
            The context size to use when running inference (in number of tokens).

        Returns
        -------
        results : dict[str, np.ndarray]
            The risk estimates for each data type in the dataset (usually "train",
            "val", "test").
        """
        data_types = {
            "train": dataset.get_train()[0],
            "test": dataset.get_test()[0],
        }
        if dataset.get_val() is not None:
            data_types["val"] = dataset.get_val()[0]

        results = {
            data_type: self._compute_risk_estimates_for_dataframe(
                df=df,
                dataset=dataset,
                batch_size=batch_size,
                context_size=context_size,
            )
            for data_type, df in data_types.items()
        }

        return results



"""Definition of a generic TaskMetadata class.
"""
from __future__ import annotations

import dataclasses
import logging
from dataclasses import dataclass, field
from typing import Callable, ClassVar, Iterable

import pandas as pd

from ._utils import hash_dict
from .col_to_text import ColumnToText


@dataclass
class TaskMetadata:
    """A base class to hold information on a prediction task.

    Attributes
    ----------
    name : str
        The name of the task.
    description : str
        A description of the task, including the population to which the task
        pertains to.
    features : list[str]
        The names of the features used in the task.
    target : str
        The name of the target column.
    cols_to_text : dict[str, ColumnToText]
        A mapping between column names and their textual descriptions.
    sensitive_attribute : str, optional
        The name of the column used as the sensitive attribute data (if provided).
    target_threshold : float, optional
        The threshold used to binarize the target column (if provided).
    """
    name: str
    description: str
    features: list[str]
    target: str
    cols_to_text: dict[str, ColumnToText]
    sensitive_attribute: str = None
    target_threshold: float | int = None

    # Class-level task storage
    _tasks: ClassVar[dict[str, "TaskMetadata"]] = field(default={}, init=False, repr=False)

    def __post_init__(self):
        # Check if this task had already been created
        if self.name in TaskMetadata._tasks:
            logging.error(f"A task with `name='{self.name}'` already exists. Overwriting...")

        # Add this task to the class-level dictionary
        TaskMetadata._tasks[self.name] = self

    def __hash__(self) -> int:
        hashable_params = dataclasses.asdict(self)
        hashable_params.pop("cols_to_text")
        hashable_params["question_hash"] = hash(self.question)
        return int(hash_dict(hashable_params), 16)

    @classmethod
    def get_task(cls, name: str) -> "TaskMetadata":
        if name not in cls._tasks:
            raise ValueError(f"Task '{name}' has not been created yet.")
        return cls._tasks[name]

    @property
    def question(self):
        if self.cols_to_text[self.target]._question is None:
            raise ValueError(f"No question provided for the target column '{self.target}'.")
        return self.cols_to_text[self.target].question

    def get_row_description(self, row: pd.Series) -> str:
        """Encode a description of a given data row in textual form."""
        row = row[self.features]
        return (
            "\n".join(
                "- " + self.cols_to_text[col].get_text(val)
                for col, val in row.items()
            )
        )

    def sensitive_attribute_value_map(self) -> Callable:
        """Returns a mapping between sensitive attribute values and their descriptions."""
        if self.sensitive_attribute is None:
            logging.warning("No sensitive attribute provided for this task.")
            return {}
        return self.cols_to_text[self.sensitive_attribute].value_map

    def create_task_with_feature_subset(self, feature_subset: Iterable[str]):
        """Creates a new task with a subset of the original features."""
        # Convert iterable to list
        feature_subset = list(feature_subset)

        # Check if features are a subset of the original features
        if not set(feature_subset).issubset(self.features):
            raise ValueError("`feature_subset` must be a subset of the original features.")

        # Return new TaskMetadata object
        return dataclasses.replace(
            self,
            name=f"{self.name}_" + "_".join(sorted(feature_subset)),
            features=feature_subset,
        )



"""Module to access ACS data using the folktables package.
"""
from __future__ import annotations

import logging
from pathlib import Path

from folktables import ACSDataSource
from folktables.load_acs import state_list

from ..dataset import Dataset
from ._utils import get_thresholded_column_name
from .acs_tasks import ACSTaskMetadata  # noqa # load ACS tasks

DEFAULT_ACS_DATA_DIR = Path("~/data").expanduser().resolve()
DEFAULT_SEED = 42

DEFAULT_SURVEY_YEAR = "2018"
DEFAULT_SURVEY_HORIZON = "1-Year"
DEFAULT_SURVEY_UNIT = "person"


class ACSDataset(Dataset):
    """Wrapper for ACS folktables datasets."""

    def __init__(
        self,
        task: str | ACSTaskMetadata,
        cache_dir: str | Path = None,
        survey_year: str = DEFAULT_SURVEY_YEAR,
        horizon: str = DEFAULT_SURVEY_HORIZON,
        survey: str = DEFAULT_SURVEY_UNIT,
        seed: int = DEFAULT_SEED,
        **kwargs,
    ):
        # Create "folktables" sub-folder under the given cache dir
        cache_dir = Path(cache_dir or DEFAULT_ACS_DATA_DIR).expanduser().resolve() / "folktables"
        if not cache_dir.exists():
            logging.warning(f"Creating cache directory '{cache_dir}' for ACS data.")
            cache_dir.mkdir(exist_ok=True, parents=False)

        # Load ACS data source
        print("Loading ACS data...")
        data_source = ACSDataSource(
            survey_year=survey_year, horizon=horizon, survey=survey,
            root_dir=cache_dir.as_posix(),
        )

        # Get ACS data in a pandas DF
        data = data_source.get_data(
            states=state_list, download=True, random_seed=seed,
        )

        # Get information on this ACS/folktables task
        task = ACSTaskMetadata.get_task(task) if isinstance(task, str) else task

        # Keep only rows used in this task
        data = task.folktables_obj._preprocess(data)

        # Threshold the target column if necessary
        # > use standardized ACS naming convention
        if task.target_threshold is not None:
            thresholded_target = get_thresholded_column_name(task.target, task.target_threshold)
            data[thresholded_target] = (data[task.target] >= task.target_threshold).astype(int)
            task.target = thresholded_target

        super().__init__(
            data=data,
            task=task,
            seed=seed,
            **kwargs,
        )



#!/usr/bin/env python3
"""
Python script to launch condor jobs for all experiments.
"""
import argparse
import math
from pathlib import Path
from pprint import pprint

from folktexts._io import load_json, save_json
from folktexts.llm_utils import get_model_folder_path, get_model_size_B

from .experiments import Experiment, launch_experiment_job

# All ACS prediction tasks
ACS_TASKS = (
    "ACSIncome",
    # "ACSEmployment",          # TODO: run on other ACS tasks
    # "ACSMobility",
    # "ACSTravelTime",
    # "ACSPublicCoverage",
)

################
# Useful paths #
################
ROOT_DIR = Path("/fast/groups/sf")
# ROOT_DIR = Path("~").expanduser().resolve()               # on local machine

# ACS data directory
ACS_DATA_DIR = ROOT_DIR / "data"

# Models save directory
MODELS_DIR = ROOT_DIR / "huggingface-models"
# MODELS_DIR = ROOT_DIR / "data" / "huggingface-models"     # on local machine

# Path to the executable script to run
EXECUTABLE_PATH = Path(__file__).parent.resolve() / "run_acs_benchmark.py"

##################
# Global configs #
##################
BATCH_SIZE = 30
CONTEXT_SIZE = 500
CORRECT_ORDER_BIAS = True

VERBOSE = True

JOB_CPUS = 4
JOB_MEMORY_GB = 60
JOB_BID = 250

# LLMs to evaluate
LLM_MODELS = [
    # ** Small models **
    "google/gemma-2b",
    "google/gemma-1.1-2b-it",

    # # ** Medium models **
    "google/gemma-7b",
    "google/gemma-1.1-7b-it",
    "mistralai/Mistral-7B-v0.1",
    "mistralai/Mistral-7B-Instruct-v0.2",
    "meta-llama/Meta-Llama-3-8B",
    "meta-llama/Meta-Llama-3-8B-Instruct",

    # # ** Large models **
    "01-ai/Yi-34B",
    "01-ai/Yi-34B-Chat",
    "mistralai/Mixtral-8x7B-v0.1",
    "mistralai/Mixtral-8x7B-Instruct-v0.1",
    "meta-llama/Meta-Llama-3-70B",
    "meta-llama/Meta-Llama-3-70B-Instruct",
    "mistralai/Mixtral-8x22B-v0.1",
    "mistralai/Mixtral-8x22B-Instruct-v0.1",
]


# Function that defines common settings among all LLM-as-clf experiments
def make_llm_as_clf_experiment(
    model_name: str,
    task_name: str,
    results_root_dir: str,
    **kwargs,
) -> Experiment:
    """Create an experiment object to run.
    """
    # Get model size
    model_size_B = get_model_size_B(model_name)

    # Get model path
    model_path = get_model_folder_path(model_name, root_dir=MODELS_DIR)
    assert Path(model_path).exists(), f"Model path '{model_path}' does not exist."

    # Split experiment and job kwargs
    job_kwargs = {key: val for key, val in kwargs.items() if key.startswith("job_")}
    experiment_kwargs = {key: val for key, val in kwargs.items() if key not in job_kwargs}

    # Set default job kwargs
    job_kwargs.setdefault("job_cpus", JOB_CPUS)
    job_kwargs.setdefault("job_gpus", math.ceil(model_size_B / 40))     # One GPU per 40B parameters
    job_kwargs.setdefault("job_memory_gb", JOB_MEMORY_GB)
    job_kwargs.setdefault("job_gpu_memory_gb", 35 if model_size_B < 5 else 60)
    job_kwargs.setdefault("job_bid", JOB_BID)

    # Set default experiment kwargs
    n_shots = int(experiment_kwargs.get("few_shot", 1))
    experiment_kwargs.setdefault("batch_size", math.ceil(BATCH_SIZE / n_shots))
    experiment_kwargs.setdefault("context_size", CONTEXT_SIZE * n_shots)
    experiment_kwargs.setdefault("data_dir", ACS_DATA_DIR.as_posix())
    # experiment_kwargs.setdefault("fit_threshold", FIT_THRESHOLD)

    # Define experiment
    exp = Experiment(
        executable_path=EXECUTABLE_PATH.as_posix(),
        kwargs=dict(
            model=model_path,
            task_name=task_name,
            **experiment_kwargs,
        ),
        **job_kwargs,
    )

    # Create LLM results directory
    exp_results_dir = Path(results_root_dir) / get_llm_results_folder(exp)
    exp_results_dir.mkdir(exist_ok=True, parents=True)
    exp.kwargs["results_dir"] = exp_results_dir.as_posix()
    save_json(
        obj=exp.to_dict(),
        path=exp_results_dir / "experiment.json",
        overwrite=True,
    )

    return exp


def get_llm_results_folder(exp: Experiment) -> str:
    """Create a unique experiment name.
    """
    return (
        f"model-{Path(exp.model).name}."
        f"dataset-{exp.task_name}"
    )


def setup_arg_parser() -> argparse.ArgumentParser:
    # Init parser
    parser = argparse.ArgumentParser(description="Launch experiments to evaluate LLMs as classifiers.")

    parser.add_argument(
        "--results-root-dir",
        type=str,
        help="[string] Directory under which results will be saved.",
        required=True,
    )

    parser.add_argument(
        "--model",
        type=str,
        help="[string] Model name on huggingface hub - can provide multiple!",
        required=False,
        action="append",
    )

    parser.add_argument(
        "--task-name",
        type=str,
        help="[string] ACS task name to run experiments on - can provide multiple!",
        required=False,
        action="append",
    )

    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Construct folder structure and print experiments without launching them.",
        default=False,
    )

    parser.add_argument(
        "--experiment-json",
        type=str,
        help="[string] Path to an experiment JSON file to load. Will override all other args.",
        required=False,
    )

    return parser


def main():
    # Parse command-line arguments
    parser = setup_arg_parser()
    args, extra_kwargs = parser.parse_known_args()

    # Parse extra kwargs
    from ._utils import cmd_line_args_to_kwargs
    extra_kwargs = cmd_line_args_to_kwargs(extra_kwargs)
    # TODO: use the run_acs_benchmark.py parser to parse extra kwargs
    #       with `setup_arg_parser().convert_arg_line_to_args(extra_kwargs)` !!!

    models = args.model or LLM_MODELS
    tasks = args.task_name or ACS_TASKS

    # Load experiment from JSON file if provided
    if args.experiment_json:
        print(f"Launching job for experiment at '{args.experiment_json}'...")
        exp = Experiment(**load_json(args.experiment_json))
        all_experiments = [exp]

    # Otherwise, run all experiments planned
    else:
        all_experiments = [
            make_llm_as_clf_experiment(
                model_name=model,
                task_name=task,
                results_root_dir=args.results_root_dir,
                **extra_kwargs,
            )
            for model in models
            for task in tasks
        ]

    # Log total number of experiments
    print(f"Launching {len(all_experiments)} experiment(s)...\n")
    for i, exp in enumerate(all_experiments):
        cluster_id = launch_experiment_job(exp).cluster() if not args.dry_run else None
        print(f"{i:2}. cluster-id={cluster_id}")
        pprint(exp.to_dict(), indent=4)


if __name__ == "__main__":
    main()



"""General constants and helper classes to run the main experiments on htcondor.
"""
import logging
import sys
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import ClassVar

import classad
import htcondor

from folktexts._utils import hash_dict

# Cluster settings
DEFAULT_JOB_BID = 25            # htcondor bid (min. is 15 apparently...)
DEFAULT_JOB_CPUS = 4            # number of CPUs per experiment (per cluster job)
DEFAULT_JOB_MEMORY_GB = 62      # GBs of memory
DEFAULT_GPU_MEMORY_GB = 30      # GBs of GPU memory

MAX_RUNNING_PRICE = 1500        # Max price for running a job


@dataclass
class Experiment:
    """A generic experiment to run on the cluster."""
    executable_path: str
    kwargs: dict = field(default_factory=dict)

    job_cpus: int = DEFAULT_JOB_CPUS
    job_gpus: int = 0
    job_memory_gb: int = DEFAULT_JOB_MEMORY_GB
    job_gpu_memory_gb: int = DEFAULT_GPU_MEMORY_GB
    job_bid: int = DEFAULT_JOB_BID

    _all_experiments: ClassVar[list["Experiment"]] = []

    def __post_init__(self):
        # Add experiment to the class-level list
        self._all_experiments.append(self)

    @classmethod
    def get_all_experiments(cls):
        return cls._all_experiments

    def __getattr__(self, name: str):
        if name in self.kwargs:
            return self.kwargs[name]
        else:
            raise AttributeError(f"Attribute '{name}' not found in Experiment.")

    def hash(self) -> str:
        """Generate a hexadecimal hash that uniquely identifies the experiment's arguments.
        """
        # Get hash of the experiment's arguments
        kwargs_for_hash = dict(
            executable_path=self.executable_path,
            **self.kwargs,
        )

        # These kwargs shouldn't be used to generate a unique hash
        kwargs_for_hash.pop("results_dir", None)
        kwargs_for_hash.pop("hash", None)

        return hash_dict(kwargs_for_hash)

    def to_dict(self) -> dict:
        return asdict(self)


def launch_experiment_job(exp: Experiment):

    # Name/prefix for cluster logs related to this job
    cluster_job_log_name = (Path(exp.results_dir) / "log.$(Cluster).$(Process)").as_posix()

    # Construct executable cmd-line arguments
    cmd_line_args = " ".join(f"--{key.replace('_', '-')} {value}" for key, value in exp.kwargs.items())

    # Construct job description
    job_description = htcondor.Submit({
        "executable": f"{sys.executable}",  # correct env for the python executable
        "arguments": f"{exp.executable_path} {cmd_line_args}",
        "output": f"{cluster_job_log_name}.out",
        "error": f"{cluster_job_log_name}.err",
        "log": f"{cluster_job_log_name}.log",
        "request_cpus": f"{exp.job_cpus}",
        "request_gpus": f"{exp.job_gpus}",
        "request_memory": f"{exp.job_memory_gb}GB",
        "request_disk": "10GB",
        "jobprio": f"{exp.job_bid - 1000}",
        "notify_user": "andre.cruz@tuebingen.mpg.de",
        "notification": "error",

        # GPU requirements
        "requirements": (
            f"(TARGET.CUDAGlobalMemoryMb > {exp.job_gpu_memory_gb * 1_000})"
        ) if exp.job_gpus > 0 else "",

        # Concurrency limits:
        # > each job uses this amount of resources out of a pool of 10k
        "concurrency_limits": "user.folktexts:100",     # 100 jobs in parallel

        "+MaxRunningPrice": MAX_RUNNING_PRICE,
        "+RunningPriceExceededAction": classad.quote("restart"),
    })

    # Submit job to the htcondor scheduler
    schedd = htcondor.Schedd()
    submit_result = schedd.submit(job_description)

    logging.info(
        f"Launched {submit_result.num_procs()} processe(s) with "
        f"cluster-ID={submit_result.cluster()}\n")

    return submit_result



"""Helper functions for ACS data processing."""
from __future__ import annotations

import logging
import re
from pathlib import Path
from typing import Callable


def get_thresholded_column_name(column_name: str, threshold: float | int) -> str:
    """Standardizes naming of thresholded columns."""
    threshold_str = f"{threshold:.2f}".replace(".", "_") if isinstance(threshold, float) else str(threshold)
    return f"{column_name}_binary_{threshold_str}"


def parse_pums_code(
    value: int,
    file: str | Path,
    postprocess: Callable[[str], str] | None = None,
    cache={},
) -> str:
    # Check if file already loaded into cache
    if file not in cache:
        line_re = re.compile(r"(?P<code>\d+)\s+[.](?P<description>.+)$")

        file_cache = {}
        with open(file) as f:
            for line in f:
                m = line_re.match(line)
                if m is None:
                    logging.error(f"Could not parse line: {line}")
                    continue

                code, description = m.group("code"), m.group("description")
                file_cache[int(code)] = postprocess(description) if postprocess else description

        cache[file] = file_cache

    # Get file cache
    file_cache = cache[file]

    # Return the value from cache, or "N/A" if not found
    if value not in file_cache:
        logging.warning(f"Could not find code '{value}' in file '{file}'")
        return "N/A"

    return file_cache[value]



"""General Dataset functionality for text-based datasets.

TODO
----
- Re-assess if the Dataset needs permanent access to the task;
    - The task is already in the LLMClassifier;
    - Maybe the Dataset should simply receive the `task` object whenever a
    method needs it.
"""
from __future__ import annotations

import copy
import logging
from abc import ABC

import numpy as np
import pandas as pd

from ._utils import hash_dict, is_valid_number
from .task import TaskMetadata

DEFAULT_TEST_SIZE = 0.1
DEFAULT_VAL_SIZE = None


class Dataset(ABC):
    def __init__(
        self,
        data: pd.DataFrame,
        task: TaskMetadata,     # TODO: remove this from the Dataset
        test_size: float = DEFAULT_TEST_SIZE,
        val_size: float = DEFAULT_VAL_SIZE,
        subsampling: float = None,
        seed: int = 42,
    ):
        """Construct a Dataset object.

        Parameters
        ----------
        data : pd.DataFrame
            The dataset's data in pandas DataFrame format.
        task : TaskMetadata
            The metadata for the prediction task.
        test_size : float, optional
            The size of the test set, as a fraction of the total dataset size,
            by default 0.1.
        val_size : float, optional
            The size of the validation set, as a fraction of the total dataset
            size, by default 0.1.
        subsampling : float, optional
            Whether to use sub-sampling, and which fraction of the data to keep.
            By default will not use sub-sampling (`subsampling=None`).
        seed : int, optional
            The random state seed, by default 42.
        """
        self._data = data
        self._task = task

        self._test_size = test_size
        self._val_size = val_size or 0
        self._train_size = 1 - self._test_size - self._val_size
        assert self._train_size > 0

        self._seed = seed
        self._rng = np.random.default_rng(self._seed)

        # Make train/test/val split
        indices = self._rng.permutation(len(self._data))
        self._train_indices = indices[: int(len(indices) * self._train_size)]
        self._test_indices = indices[
            len(self._train_indices):
            int(len(indices) * (self._train_size + self._test_size))]

        if val_size is not None and val_size > 0:
            self._val_indices = indices[
                len(self._train_indices) + len(self._test_indices):]
        else:
            self._val_indices = None

        # Subsample the train/test/val data (if requested)
        self._subsampling = None
        if subsampling is not None:
            self._subsample_inplace(subsampling)

    @property
    def data(self) -> pd.DataFrame:
        return self._data

    @property
    def task(self) -> TaskMetadata:
        return self._task

    @task.setter
    def task(self, task: TaskMetadata):
        logging.info(f"Updating dataset's task from '{self.task.name}' to '{task.name}'.")
        # Check if task columns are in the data
        if not all(col in self.data.columns for col in (task.features + [task.target])):
            raise ValueError(
                f"Task columns not found in dataset: "
                f"features={task.features}, target={task.target}")

        self._task = task

    @property
    def train_size(self) -> float:
        return self._train_size

    @property
    def test_size(self) -> float:
        return self._test_size

    @property
    def val_size(self) -> float:
        return self._val_size

    @property
    def subsampling(self) -> float:
        return getattr(self, "_subsampling", None)

    @property
    def seed(self) -> int:
        return self._seed

    @property
    def name(self) -> str:
        """A unique name for this dataset."""
        subsampling_str = f"subsampled-{self.subsampling:.3}" if self.subsampling else "full"
        seed_str = f"seed-{self._seed}"
        hash_str = f"hash-{hash(self)}"
        return f"{self.task.name}_{subsampling_str}_{seed_str}_{hash_str}"

    def __copy__(self) -> "Dataset":
        dataset = Dataset(
            data=self.data,
            task=self.task,
            test_size=self.test_size,
            val_size=self.val_size,
            subsampling=self.subsampling,
            seed=self.seed,
        )
        dataset._train_indices = self._train_indices.copy()
        dataset._test_indices = self._test_indices.copy()
        dataset._val_indices = self._val_indices.copy() if self._val_indices is not None else None
        dataset._rng = copy.deepcopy(self._rng)

        return dataset

    def _subsample_inplace(self, subsampling: float) -> "Dataset":
        """Subsample the dataset in-place."""

        # Check argument is valid
        if not is_valid_number(subsampling) or not (0 < subsampling <= 1):
            raise ValueError(f"`subsampling={subsampling}` must be in the range (0, 1]")

        # Update train/test/val indices
        new_train_size = int(len(self._train_indices) * subsampling + 0.5)
        new_test_size = int(len(self._test_indices) * subsampling + 0.5)

        self._train_indices = self._train_indices[: new_train_size]
        self._test_indices = self._test_indices[: new_test_size]
        if self._val_indices is not None:
            new_val_size = int(len(self._val_indices) * subsampling + 0.5)
            self._val_indices = self._val_indices[: new_val_size]

        # Update subsampling factor
        self._subsampling = (getattr(self, "_subsampling", None) or 1) * subsampling

        # Log new dataset size
        msg = (
            f"Subsampled dataset to {self.subsampling:.1%} of the original size. "
            f"Train size: {len(self._train_indices)}, "
            f"Test size: {len(self._test_indices)}, "
            f"Val size: {len(self._val_indices) if self._val_indices is not None else 0};"
        )
        logging.info(msg)

        return self

    def subsample(self, subsampling: float) -> "Dataset":
        """Create a new dataset whose samples are a fraction of this dataset."""
        return copy.copy(self)._subsample_inplace(subsampling)

    def _filter_inplace(
        self,
        population_feature_values: dict,
    ) -> "Dataset":
        """Subset the dataset in-place: keep only samples with the given feature values."""
        # Check argument is of valid type
        if not isinstance(population_feature_values, dict):
            raise ValueError(
                f"Invalid `population_feature_values` type: "
                f"{type(population_feature_values)}.")

        # Check argument keys are valid columns
        if not all(key in self.data.columns for key in population_feature_values.keys()):
            raise ValueError(
                f"Invalid `population_feature_values` keys; columns don't exist "
                f"in the dataset: {list(population_feature_values.keys())}.")

        # Create boolean filter based on the given feature values
        population_filter = pd.Series(True, index=self.data.index)
        for key, value in population_feature_values.items():
            population_filter &= (self.data[key] == value)

        # Update train/test/val indices
        train_pop_filter = population_filter.iloc[self._train_indices]
        test_pop_filter = population_filter.iloc[self._test_indices]
        val_pop_filter = population_filter.iloc[self._val_indices] if self._val_indices is not None else None

        self._train_indices = self._train_indices[train_pop_filter]
        self._test_indices = self._test_indices[test_pop_filter]
        self._val_indices = self._val_indices[val_pop_filter] if self._val_indices is not None else None

        return self

    def filter(self, population_feature_values: dict) -> "Dataset":
        """Create a new dataset whose samples are a subset of this dataset."""
        return copy.copy(self)._filter_inplace(population_feature_values)

    def get_features_data(self) -> pd.DataFrame:
        return self.data[self.task.features]

    def get_target_data(self) -> pd.Series:
        return self.data[self.task.target]

    def get_sensitive_attribute_data(self) -> pd.Series:
        if self.task.sensitive_attribute is not None:
            return self.data[self.task.sensitive_attribute]
        return None

    def get_data_split(self, split: str) -> tuple[pd.DataFrame, pd.Series]:
        if split == "train":
            return self.get_train()
        elif split == "test":
            return self.get_test()
        elif split == "val":
            return self.get_val()
        else:
            raise ValueError(f"Invalid split '{split}'")

    def get_train(self):
        train_data = self.data.iloc[self._train_indices]
        return train_data[self.task.features], train_data[self.task.target]

    def sample_n_train_examples(
        self,
        n: int,
        reuse_examples: bool = False,
    ) -> tuple[pd.DataFrame, pd.Series]:
        """Return a set of samples from the training set.

        Parameters
        ----------
        n : int
            The number of example rows to return.
        reuse_examples : bool, optional
            Whether to reuse the same examples for consistency. By default will
            sample new examples each time (`reuse_examples=False`).

        Returns
        -------
        X, y : tuple[pd.DataFrame, pd.Series]
            The features and target data for the sampled examples.
        """
        # TODO: make sure examples are class-balanced?
        if reuse_examples:
            example_indices = self._train_indices[:n]
        else:
            example_indices = self._rng.choice(self._train_indices, size=n, replace=False)

        return (
            self.data.iloc[example_indices][self.task.features],
            self.data.iloc[example_indices][self.task.target],
        )

    def get_test(self):
        test_data = self.data.iloc[self._test_indices]
        return test_data[self.task.features], test_data[self.task.target]

    def get_val(self):
        if self._val_indices is None:
            return None
        val_data = self.data.iloc[self._val_indices]
        return val_data[self.task.features], val_data[self.task.target]

    def __getitem__(self, i) -> tuple[pd.DataFrame, pd.Series]:
        """Returns the i-th training sample."""
        curr_indices = self._train_indices[i]
        curr_data = self.data.iloc[curr_indices]
        return curr_data[self.task.features], curr_data[self.task.target]

    def __iter__(self):
        """Iterates over the training data."""
        for i in range(len(self._train_indices)):
            yield self[i]

    def __len__(self) -> int:
        return len(self.data)

    def __hash__(self) -> int:
        hashable_params = {
            "data_shape": self.data.shape,
            "task": hash(self.task),
            "train_size": len(self._train_indices),
            "test_size": len(self._test_indices),
            "val_size": len(self._val_indices) if self._val_indices is not None else 0,
            "subsampling": self.subsampling or 1,
            "seed": self.seed,
        }

        return int(hash_dict(hashable_params), 16)



from __future__ import annotations

import json
import logging
import pickle
from collections.abc import Collection
from pathlib import Path

import cloudpickle


def save_json(obj: Collection, path: str | Path, overwrite: bool = True):
    """Saves a JSON-serializable object to disk."""
    logging.info(f"Saving JSON file to '{str(path)}'")
    with open(path, "w" if overwrite else "x") as f_out:
        json.dump(obj, f_out, indent=4, sort_keys=True)


def load_json(path: str | Path) -> object:
    """Loads a JSON file from disk and returns the deserialized object."""
    with open(path, "r") as f_in:
        return json.load(f_in)


def save_pickle(obj: object, path: str | Path, overwrite: bool = True) -> bool:
    """Saves the given object as a pickle with the given file path.

    Parameters
    ----------
    obj : object
        The object to pickle
    path : str or Path
        The file path to save the pickle with.

    Returns
    -------
    success : bool
        Whether pickling was successful.
    """
    logging.info(f"Saving pickle file to '{str(path)}'")
    try:
        with open(path, "wb" if overwrite else "xb") as f_out:
            cloudpickle.dump(obj, f_out)
            return True

    except Exception as e:
        logging.error(f"Pickling failed with exception '{e}'")
        return False


def load_pickle(path: str | Path) -> object:
    with open(path, "rb") as f_in:
        return pickle.load(f_in)



"""A benchmark class for measuring and evaluating LLM calibration.
"""
from __future__ import annotations

import dataclasses
import logging
from functools import partial
from pathlib import Path

import numpy as np
from transformers import AutoModelForCausalLM, AutoTokenizer

from ._io import load_json, save_json
from ._utils import hash_dict, is_valid_number
from .acs.acs_dataset import ACSDataset
from .acs.acs_questions import acs_multiple_choice_qa_map, acs_numeric_qa_map
from .acs.acs_tasks import ACSTaskMetadata
from .classifier import LLMClassifier
from .dataset import Dataset
from .evaluation import evaluate_predictions
from .plotting import render_evaluation_plots, render_fairness_plots
from .prompting import encode_row_prompt, encode_row_prompt_chat, encode_row_prompt_few_shot
from .task import TaskMetadata

DEFAULT_SEED = 42
DEFAULT_FIT_THRESHOLD_N = 100


@dataclasses.dataclass(frozen=True, eq=True)
class BenchmarkConfig:
    """A dataclass to hold the configuration for a calibration benchmark."""

    chat_prompt: bool = False
    direct_risk_prompting: bool = False
    few_shot: int | None = None
    reuse_few_shot_examples: bool = False
    batch_size: int | None = None
    context_size: int | None = None
    correct_order_bias: bool = True
    feature_subset: list[str] | None = None
    population_filter: dict | None = None
    seed: int = DEFAULT_SEED

    @classmethod
    def default_config(cls, **changes):
        """Returns the default configuration with optional changes."""
        return cls(**changes)

    @classmethod
    def load_from_disk(cls, path: str | Path):
        """Load the configuration from disk."""
        return cls(**load_json(path))

    def save_to_disk(self, path: str | Path):
        """Save the configuration to disk."""
        save_json(dataclasses.asdict(self), path)

    def __hash__(self) -> int:
        """Generates a unique hash for the configuration."""
        cfg = dataclasses.asdict(self)
        cfg["feature_subset"] = tuple(cfg["feature_subset"]) if cfg["feature_subset"] else None
        cfg["population_filter_hash"] = (
            hash_dict(cfg["population_filter"])
            if cfg["population_filter"] else None
        )
        return int(hash_dict(cfg), 16)


class CalibrationBenchmark:
    """A benchmark class for measuring and evaluating LLM calibration."""

    DEFAULT_BENCHMARK_METRIC = "ece"

    """
    Standardized configurations for the ACS data to use for benchmarking.
    """
    ACS_DATASET_CONFIGS = {
        # ACS survey configs
        "survey_year": "2018",
        "horizon": "1-Year",
        "survey": "person",

        # Data split configs
        "test_size": 0.1,
        "val_size": 0.1,
        "subsampling": None,

        # Fixed random seed
        "seed": 42,
    }

    def __init__(
        self,
        llm_clf: LLMClassifier,
        dataset: Dataset | str,
        config: BenchmarkConfig,
    ):
        self.llm_clf = llm_clf
        self.dataset = dataset
        self.config = config

        self._y_test_scores: np.ndarray = None
        self._results_root_dir = "."  # Default root results directory
        self._results: dict = None
        self._plots: dict = None

        # Log initialization
        msg = (
            f"\n** Benchmark initialization **\n"
            f"Model: {self.model_name};\n"
            f"Task: {self.task.name};\n"
            f"Hash: {hash(self)};\n"
        )
        logging.info(msg)

    @property
    def configs_dict(self) -> dict:
        cnf = dataclasses.asdict(self.config)

        # Add info on model, task, and dataset
        cnf["model_name"] = self.model_name
        cnf["model_hash"] = hash(self.llm_clf)
        cnf["task_name"] = self.task.name
        cnf["task_hash"] = hash(self.task)
        cnf["dataset_name"] = self.dataset.name
        cnf["dataset_hash"] = hash(self.dataset)

        return cnf

    @property
    def results(self):
        # Add benchmark configs to the results
        self._results["config"] = self.configs_dict
        return self._results

    @property
    def task(self):
        return self.llm_clf.task

    @property
    def model_name(self):
        return self.llm_clf.model_name

    @property
    def results_root_dir(self) -> Path:
        return Path(self._results_root_dir)

    @results_root_dir.setter
    def results_root_dir(self, path: str | Path):
        self._results_root_dir = Path(path).expanduser().resolve()

    @property
    def results_dir(self) -> Path:
        """Get the results directory for this benchmark."""
        # Get subfolder name
        subfolder_name = f"{self.model_name}_bench-{hash(self)}"
        subfolder_dir_path = Path(self.results_root_dir) / subfolder_name

        # Create subfolder directory if it does not exist
        subfolder_dir_path.mkdir(exist_ok=True, parents=True)
        return subfolder_dir_path

    def __hash__(self) -> int:
        hash_params = dict(
            llm_clf_hash=hash(self.llm_clf),
            dataset_hash=hash(self.dataset),
            config_hash=hash(self.config),
        )

        return int(hash_dict(hash_params), 16)

    def _get_predictions_save_path(self, data_split: str) -> Path:
        """Standardized path to file containing predictions for the given data split."""
        assert data_split in ("train", "val", "test")
        return self.results_dir / f"{self.dataset.name}.{data_split}_predictions.csv"

    def run(self, results_root_dir: str | Path, fit_threshold: int | bool = 0) -> float:
        """Run the calibration benchmark experiment.

        Parameters
        ----------
        results_root_dir : str | Path
            Path to root directory under which results will be saved.
        fit_threshold : int | bool, optional
            Whether to fit the binarization threshold on a given number of
            training samples, by default 0 (will not fit the threshold).

        Returns
        -------
        float
            The benchmark metric value. By default this is the ECE score.
        """
        if self._results is not None:
            logging.warning("Benchmark was already run. Overriding previous results.")

        # Update results directory
        self.results_root_dir = results_root_dir

        # Get test data
        X_test, y_test = self.dataset.get_test()
        logging.info(f"Test data features shape: {X_test.shape}")

        # Get sensitive attribute data if available
        s_test = None
        if self.task.sensitive_attribute is not None:
            s_test = self.dataset.get_sensitive_attribute_data().loc[y_test.index]

        # Get LLM risk-estimate predictions for each row in the test set
        test_predictions_save_path = self._get_predictions_save_path("test")
        self._y_test_scores = self.llm_clf.predict_proba(
            data=X_test,
            predictions_save_path=test_predictions_save_path,
            labels=y_test,  # used only to save alongside predictions in disk
        )

        # If requested, fit the threshold on a small portion of the train set
        if fit_threshold:
            if fit_threshold is True:
                fit_threshold = DEFAULT_FIT_THRESHOLD_N
            elif not is_valid_number(fit_threshold) or fit_threshold <= 0:
                raise ValueError(f"Invalid fit_threshold={fit_threshold}")

            logging.info(f"Fitting threshold on {fit_threshold} train samples")
            X_train, y_train = self.dataset.sample_n_train_examples(fit_threshold)
            self.llm_clf.fit(X_train, y_train)

        # Evaluate test risk scores
        self._results = evaluate_predictions(
            y_true=y_test.to_numpy(),
            y_pred_scores=self._y_test_scores,
            sensitive_attribute=s_test,
            threshold=self.llm_clf.threshold,
            model_name=self.llm_clf.model_name,
        )

        # Save predictions save path
        self._results["predictions_path"] = test_predictions_save_path.as_posix()

        # Log main results
        msg = (
            f"\n** Test results **\n"
            f"Model: {self.llm_clf.model_name};\n"
            f"\t ECE:       {self._results['ece']:.1%};\n"
            f"\t ROC AUC :  {self.results['roc_auc']:.1%};\n"
            f"\t Accuracy:  {self.results['accuracy']:.1%};\n"
            f"\t Bal. acc.: {self.results['balanced_accuracy']:.1%};\n"
        )
        logging.info(msg)

        # Render plots
        self.plot_results(show_plots=False)

        # Save results to disk
        self.save_results()

        return self._results[self.DEFAULT_BENCHMARK_METRIC]

    def plot_results(self, *, show_plots: bool = True):
        """Render evaluation plots and save to disk.

        Parameters
        ----------
        show_plots : bool, optional
            Whether to show plots, by default True.

        Returns
        -------
        plots_paths : dict[str, str]
            The paths to the saved plots.
        """
        if self._results is None:
            raise ValueError("No results to plot. Run the benchmark first.")

        imgs_dir = Path(self.results_dir) / "imgs"
        imgs_dir.mkdir(exist_ok=True, parents=False)
        _, y_test = self.dataset.get_test()

        plots_paths = render_evaluation_plots(
            y_true=y_test.to_numpy(),
            y_pred_scores=self._y_test_scores,
            eval_results=self.results,
            model_name=self.llm_clf.model_name,
            imgs_dir=imgs_dir,
            show_plots=show_plots,
        )

        # Plot fairness plots if sensitive attribute is provided
        if self.task.sensitive_attribute is not None:
            s_test = self.dataset.get_sensitive_attribute_data().loc[y_test.index]

            plots_paths.update(render_fairness_plots(
                y_true=y_test.to_numpy(),
                y_pred_scores=self._y_test_scores,
                sensitive_attribute=s_test,
                eval_results=self.results,
                model_name=self.llm_clf.model_name,
                group_value_map=self.task.sensitive_attribute_value_map(),
                imgs_dir=imgs_dir,
                show_plots=show_plots,
            ))

        self._results["plots"] = plots_paths

        return plots_paths

    def save_results(self, results_root_dir: str | Path = None):
        """Save the benchmark results to disk.

        Parameters
        ----------
        results_root_dir : str | Path, optional
            Path to root directory under which results will be saved. By default
            will use `self.results_root_dir`.
        """
        if self.results is None:
            raise ValueError("No results to save. Run the benchmark first.")

        # Update results directory if provided
        if results_root_dir is not None:
            self.results_root_dir = results_root_dir

        # Save results to disk
        results_file_name = f"results.bench-{hash(self)}.json"
        results_file_path = self.results_dir / results_file_name

        save_json(self.results, path=results_file_path)
        logging.info(f"Saved experiment results to '{results_file_path.as_posix()}'")

    @classmethod
    def make_acs_benchmark(
        cls,
        model: AutoModelForCausalLM,
        tokenizer: AutoTokenizer,
        task_name: str,
        data_dir: str | Path = None,
        config: BenchmarkConfig = BenchmarkConfig.default_config(),
        **kwargs,
    ) -> CalibrationBenchmark:
        """Create a standardized calibration benchmark on ACS data.

        Parameters
        ----------
        model : AutoModelForCausalLM
            The torch/transformers language model to use.
        tokenizer : AutoTokenizer
            The tokenizer used to train the model.
        task_name : str
            The name of the ACS task to use.
        data_dir : str | Path, optional
            Path to the directory to load data from and save data in.
        config : BenchmarkConfig, optional
            Extra benchmark configurations, by default will use
            `BenchmarkConfig.default_config()`.
        **kwargs
            Additional arguments passed to `ACSDataset`. By default will use a
            set of standardized dataset configurations for reproducibility.

        Returns
        -------
        bench : CalibrationBenchmark
            The ACS calibration benchmark object.
        """

        # Handle non-standard ACS arguments
        acs_dataset_configs = cls.ACS_DATASET_CONFIGS.copy()
        for arg in acs_dataset_configs:
            if arg in kwargs and kwargs[arg] != cls.ACS_DATASET_CONFIGS[arg]:
                logging.warning(
                    f"Received non-standard ACS argument '{arg}' (using "
                    f"{arg}={kwargs[arg]} instead of default {arg}={cls.ACS_DATASET_CONFIGS[arg]}). "
                    f"This may affect reproducibility.")
                acs_dataset_configs[arg] = kwargs.pop(arg)

        # Log unused kwargs
        if kwargs:
            logging.warning(f"Unused key-word arguments: {kwargs}")

        # Fetch ACS task and dataset
        acs_task = ACSTaskMetadata.get_task(task_name)
        acs_dataset = ACSDataset(
            task=acs_task,
            cache_dir=data_dir,
            **acs_dataset_configs)

        return cls.make_benchmark(
            model=model,
            tokenizer=tokenizer,
            task=acs_task,
            dataset=acs_dataset,
            config=config,
        )

    @classmethod
    def make_benchmark(
        cls,
        model: AutoModelForCausalLM,
        tokenizer: AutoTokenizer,
        task: TaskMetadata | str,
        dataset: Dataset,
        config: BenchmarkConfig = BenchmarkConfig.default_config(),
    ) -> CalibrationBenchmark:
        """Create a calibration benchmark from a given configuration.

        Parameters
        ----------
        model : AutoModelForCausalLM
            The torch/transformers language model to use.
        tokenizer : AutoTokenizer
            The tokenizer used to train the model.
        task : TaskMetadata | str
            The task metadata object or name of the task to use.
        dataset : Dataset
            The dataset to use for the benchmark.
        config : BenchmarkConfig, optional
            Extra benchmark configurations, by default will use
            `BenchmarkConfig.default_config()`.

        Returns
        -------
        bench : CalibrationBenchmark
            The calibration benchmark object.
        """
        # Handle TaskMetadata object
        if isinstance(task, str):
            task = TaskMetadata.get_task(task)

        if config.feature_subset is not None and len(config.feature_subset) > 0:
            task = task.create_task_with_feature_subset(config.feature_subset)
            dataset.task = task

        # Check dataset is compatible with task
        if dataset.task is not task and dataset.task.name != task.name:
            raise ValueError(
                f"Dataset task '{dataset.task.name}' does not match the "
                f"provided task '{task.name}'.")

        if config.population_filter is not None:
            dataset = dataset.filter(config.population_filter)

        # Get prompting function
        if config.chat_prompt:
            logging.warning(f"Untested feature: chat_prompt={config.chat_prompt}")  # TODO!
            encode_row_function = partial(encode_row_prompt_chat, task=task, tokenizer=tokenizer)
        else:
            encode_row_function = partial(encode_row_prompt, task=task)

        if config.few_shot:
            assert not config.chat_prompt, "Few-shot prompting is not currently compatible with chat prompting."
            encode_row_function = partial(
                encode_row_prompt_few_shot,
                task=task,
                n_shots=config.few_shot,
                dataset=dataset,
                reuse_examples=config.reuse_few_shot_examples,
            )

        # Load the QA interface to be used for risk-score prompting
        if config.direct_risk_prompting:
            logging.warning(f"Untested feature: direct_risk_prompting={config.direct_risk_prompting}")  # TODO!
            question = acs_numeric_qa_map[task.target]
        else:
            question = acs_multiple_choice_qa_map[task.target]

        # Set the task's target question
        task.cols_to_text[task.target]._question = question

        # Construct the LLMClassifier object
        llm_inference_kwargs = {"correct_order_bias": config.correct_order_bias}
        if config.batch_size:
            llm_inference_kwargs["batch_size"] = config.batch_size
        if config.context_size:
            llm_inference_kwargs["context_size"] = config.context_size

        llm_clf = LLMClassifier(
            model=model,
            tokenizer=tokenizer,
            task=task,
            encode_row=encode_row_function,
            **llm_inference_kwargs,
        )

        return cls(
            llm_clf=llm_clf,
            dataset=dataset,
            config=config,
        )



#!/usr/bin/env python3
"""Runs the LLM calibration benchmark from the command line.
"""
import json
import logging
import sys
from argparse import ArgumentParser
from pathlib import Path

DEFAULT_BATCH_SIZE = 30
DEFAULT_CONTEXT_SIZE = 500
DEFAULT_SEED = 42


def setup_arg_parser() -> ArgumentParser:

    # Init parser
    parser = ArgumentParser(description="Run an LLM as a classifier experiment.")

    # List of command-line arguments, with type and helper string
    cli_args = [
        ("--model",         str, "[str] Model name or path to model saved on disk"),
        ("--task-name",     str, "[str] Name of the ACS task to run the experiment on"),
        ("--results-dir",   str, "[str] Directory under which this experiment's results will be saved"),
        ("--data-dir",      str, "[str] Root folder to find datasets on"),
        ("--few-shot",      int, "[int] Use few-shot prompting with the given number of shots", False),
        ("--batch-size",    int, "[int] The batch size to use for inference", False, DEFAULT_BATCH_SIZE),
        ("--context-size",  int, "[int] The maximum context size when prompting the LLM", False, DEFAULT_CONTEXT_SIZE),
        ("--fit-threshold", int, "[int] Whether to fit the prediction threshold, and on how many samples", False),
        ("--subsampling",   float, "[float] Which fraction of the dataset to use (if omitted will use all data)", False),
        ("--seed",          int, "[int] Random seed -- to set for reproducibility", False, DEFAULT_SEED),
    ]

    for arg in cli_args:
        parser.add_argument(
            arg[0],
            type=arg[1],
            help=arg[2],
            required=(arg[3] if len(arg) > 3 else True),    # NOTE: required by default
            default=(arg[4] if len(arg) > 4 else None),     # default value if provided
        )

    # Add special arguments (e.g., boolean flags or multiple-choice args)
    parser.add_argument(
        "--dont-correct-order-bias",
        help="[bool] Whether to avoid correcting ordering bias, by default will correct it",
        action="store_false",
        default=True,
    )

    parser.add_argument(
        "--chat-prompt",
        help="[bool] Whether to use chat-based prompting (for instruct models)",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "--direct-risk-prompting",
        help="[bool] Whether to directly prompt for risk-estimates instead of multiple-choice Q&A",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "--reuse-few-shot-examples",
        help="[bool] Whether to reuse the same samples for few-shot prompting (or sample new ones every time)",
        action="store_true",
        default=False,
    )

    # Optionally, receive a list of features to use (subset of original list)
    parser.add_argument(
        "--use-feature-subset",
        type=str,
        nargs="*",
        help="[str] Optional subset of features to use for prediction",
        required=False,
    )

    parser.add_argument(
        "--use-population-filter",
        type=str,
        nargs="*",
        help=(
            "[str] Optional population filter for this benchmark; must follow "
            "the format 'column_name=value' to filter the dataset by a specific value."
        ),
        required=False,
    )

    parser.add_argument(
        "--logger-level",
        type=str,
        help="[str] The logging level to use for the experiment",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        required=False,
    )

    return parser


def main():
    """Prepare and launch the LLM-as-classifier experiment using ACS data."""

    # Setup parser and process cmd-line args
    parser = setup_arg_parser()
    args = parser.parse_args()

    logging.getLogger().setLevel(level=args.logger_level or "INFO")
    pretty_args_str = json.dumps(vars(args), indent=4, sort_keys=True)
    logging.info(f"Current python executable: '{sys.executable}'")
    logging.info(f"Received the following cmd-line args: {pretty_args_str}")

    # Parse population filter if provided
    population_filter_dict = None
    if args.use_population_filter:
        from ._utils import cmd_line_args_to_kwargs
        population_filter_dict = cmd_line_args_to_kwargs(args.use_population_filter)

    # Load model and tokenizer
    from folktexts.llm_utils import load_model_tokenizer
    model, tokenizer = load_model_tokenizer(args.model)

    # Fill ACS Benchmark config
    from folktexts.benchmark import BenchmarkConfig
    config = BenchmarkConfig(
        few_shot=args.few_shot,
        chat_prompt=args.chat_prompt,
        direct_risk_prompting=args.direct_risk_prompting,
        reuse_few_shot_examples=args.reuse_few_shot_examples,
        batch_size=args.batch_size,
        context_size=args.context_size,
        correct_order_bias=not args.dont_correct_order_bias,
        feature_subset=args.use_feature_subset or None,
        population_filter=population_filter_dict,
        seed=args.seed,
    )

    # Create ACS Benchmark object
    from folktexts.benchmark import CalibrationBenchmark
    bench = CalibrationBenchmark.make_acs_benchmark(
        model=model,
        tokenizer=tokenizer,
        task_name=args.task_name,
        data_dir=args.data_dir,
        config=config,
        subsampling=args.subsampling,
    )

    # Create results directory if needed
    results_dir = Path(args.results_dir).expanduser().resolve()
    results_dir.mkdir(parents=False, exist_ok=True)

    # Run benchmark
    bench.run(results_root_dir=results_dir, fit_threshold=args.fit_threshold)
    bench.save_results()

    # Save results
    import pprint
    pprint.pprint(bench.results, indent=4, sort_dicts=True)

    # Finish
    from folktexts._utils import get_current_timestamp
    print(f"\nFinished experiment successfully at {get_current_timestamp()}\n")


if __name__ == "__main__":
    main()



"""Module to hold ACS column mappings from values to natural text.
"""
from functools import partial
from pathlib import Path

from ..col_to_text import ColumnToText
from ..qa_interface import Choice, DirectNumericQA, MultipleChoiceQA
from ._utils import get_thresholded_column_name, parse_pums_code

# Path to ACS codebook files
ACS_OCCP_FILE = Path(__file__).parent / "data" / "OCCP-codes-acs.txt"
ACS_POBP_FILE = Path(__file__).parent / "data" / "POBP-codes-acs.txt"


# Describing ACS columns and corresponding questions
acs_age = ColumnToText(
    "AGEP",
    short_description="age",
    value_map=lambda x: f"{int(x)} years old",
)

acs_class_of_worker = ColumnToText(
    "COW",
    short_description="current employment status",
    question=MultipleChoiceQA(
        column="COW",
        text=(
            "Which one of the following best describes this person's employment "
            "last week or the most recent employment in the past 5 years?",
        ),
        choices=[
            Choice("working for a for-profit private company or organization", 1),
            Choice("working for a non-profit organization", 2),
            Choice("working for the local government", 3),
            Choice("working for the state government", 4),
            Choice("working for the federal government", 5),
            Choice("owner of non-incorporated business, professional practice, or farm", 6),
            Choice("owner of incorporated business, professional practice, or farm", 7),
            Choice("working without pay in a for-profit family business or farm", 8),
            Choice("unemployed and last worked 5 years ago or earlier or never worked", 9),
        ]
    ),
)

acs_schooling = ColumnToText(
    "SCHL",
    short_description="highest grade completed",
    question=MultipleChoiceQA(
        column="SCHL",
        text="What is this person's highest grade or level of school completed?",
        choices=[
            Choice("N/A - no schooling completed", 1),
            Choice("nursery school / preschool", 2),
            Choice("kindergarten", 3),
            Choice("1st grade only", 4),
            Choice("2nd grade", 5),
            Choice("3rd grade", 6),
            Choice("4th grade", 7),
            Choice("5th grade", 8),
            Choice("6th grade", 9),
            Choice("7th grade", 10),
            Choice("8th grade", 11),
            Choice("9th grade", 12),
            Choice("10th grade", 13),
            Choice("11th grade", 14),
            Choice("12th grade, no diploma", 15),
            Choice("regular high school diploma", 16),
            Choice("GED or alternative credential", 17),
            Choice("some college, less than 1 year", 18),
            Choice("some college, 1 or more years, no degree", 19),
            Choice("Associate's degree", 20),
            Choice("Bachelor's degree", 21),
            Choice("Master's degree", 22),
            Choice("Professional degree beyond a bachelor's degree", 23),
            Choice("Doctorate degree", 24),
        ],
    ),
)

acs_marital_status = ColumnToText(
    "MAR",
    short_description="marital status",
    value_map={
        1: "married",
        2: "widowed",
        3: "divorced",
        4: "separated",
        5: "never married",
    },
)

acs_occupation = ColumnToText(
    "OCCP",
    short_description="occupation",
    value_map=partial(
        parse_pums_code,
        file=ACS_OCCP_FILE,
        postprocess=lambda x: x[4:].lower().strip(),
    ),
)

acs_place_of_birth = ColumnToText(
    "POBP",
    short_description="place of birth",
    value_map=partial(parse_pums_code, file=ACS_POBP_FILE),
)

acs_relationship = ColumnToText(
    "RELP",
    short_description="relationship to the reference person in the household",
    value_map={
        0: "the 'reference person' itself",
        1: "husband/wife",
        2: "biological son or daughter",
        3: "adopted son or daughter",
        4: "stepson or stepdaughter",
        5: "brother or sister",
        6: "father or mother",
        7: "grandchild",
        8: "parent-in-law",
        9: "son-in-law or daughter-in-law",
        10: "other relative",
        11: "roomer or boarder",
        12: "housemate or roommate",
        13: "unmarried partner",
        14: "foster child",
        15: "other non-relative",
        16: "institutionalized group quarters population",
        17: "non-institutionalized group quarters population",
    },
)

acs_work_hours = ColumnToText(
    "WKHP",
    short_description="usual number of hours worked per week",
    missing_value_fill="N/A (less than 16 years old, or did not work during the past 12 months)",
    value_map=lambda x: f"{int(x)} hours",
)

acs_sex = ColumnToText(
    "SEX",
    short_description="sex",
    value_map={
        1: "Male",
        2: "Female",
    },
)

acs_race = ColumnToText(
    "RAC1P",
    short_description="race",
    value_map={
        1: "White",
        2: "Black or African American",
        3: "American Indian",
        4: "Alaska Native",
        5: (
            "American Indian and Alaska Native tribes specified, or American "
            "Indian or Alaska Native, not specified and no other races"),
        6: "Asian",
        7: "Native Hawaiian and Other Pacific Islander",
        # 8: "Some other race alone",
        8: "Some other race alone (non-White)",
        9: "Two or more races",
    },
)

acs_income = ColumnToText(
    "PINCP",
    short_description="yearly income",
    missing_value_fill="N/A (less than 15 years old)",
    value_map=lambda x: f"${int(x):,}",
)

acs_income_binary_qa = MultipleChoiceQA(
    column=get_thresholded_column_name("PINCP", 50_000),
    text="What is this person's estimated yearly income?",
    choices=[
        Choice("Below $50,000", 0),
        Choice("Above $50,000", 1),
    ],
)

acs_income_numeric_qa = DirectNumericQA(
    column=get_thresholded_column_name("PINCP", 50_000),
    text=(
        "What is the probability that this person's estimated yearly income is "
        "above $50,000 ?"
    ),
    answer_probability=True,
    num_forward_passes=2,
)

acs_income = ColumnToText(
    name=get_thresholded_column_name("PINCP", 50_000),
    short_description="yearly income",
    missing_value_fill="N/A (less than 15 years old)",
    question=acs_income_binary_qa,
)

acs_income_brackets = ColumnToText(
    "PINCP_brackets",
    short_description="yearly income",
    missing_value_fill="N/A (less than 15 years old)",
    question=MultipleChoiceQA(
        column="PINCP_brackets",
        text="What is this person's estimated yearly income?",
        choices=[
            Choice("Less than $25,000", data_value="(0.0, 25000.0]", numeric_value=12_500),
            Choice("Between $25,000 and $50,000", data_value="(25000.0, 50000.0]", numeric_value=37_500),
            Choice("Between $50,000 and $100,000", data_value="(50000.0, 100000.0]", numeric_value=75_000),
            Choice("Above $100,000", data_value="(100000.0, inf]", numeric_value=150_000),
        ],
    ),
)

acs_pubcov = ColumnToText(
    "PUBCOV",
    short_description="public health insurance coverage",
    question=MultipleChoiceQA(
        column="PUBCOV",
        text="Does this person have public health insurance coverage?",
        choices=[
            Choice("No, individual is not covered by public health insurance", 0),
            Choice("Yes, individual is covered by public health insurance", 1),
        ],
    ),
)



from __future__ import annotations

import logging
from typing import Callable

from .qa_interface import MultipleChoiceQA, QAInterface


class ColumnToText:
    """Maps a single column's values to natural text."""

    def __init__(
        self,
        name: str,
        short_description: str,
        value_map: dict[object, str] | Callable = None,
        question: QAInterface = None,
        connector_verb: str = "is",
        missing_value_fill: str = "N/A",
    ):
        """Constructs a `ColumnToText` object.

        Parameters
        ----------
        name : str
            The column's name.
        short_description : str
            A short description of the column to be used before different
            values. For example, short_description="yearly income" will result
            in "The yearly income is [...]".
        value_map : dict[int | str, str] | Callable, optional
            A map between column values and their textual meaning. If not
            provided, will try to infer a mapping from the `question`.
        question : QAInterface, optional
            A question associated with the column. If not provided, will try to
            infer a multiple-choice question from the `value_map`.
        connector_verb : str, optional
            Which verb to use when connecting the column's description to its
            value; by default "is".
        missing_value_fill : str, optional
            The value to use when the column's value is not found in the
            `value_map`, by default "N/A".
        """
        self._name = name
        self._short_description = short_description
        self._value_map = value_map
        self._question = question
        self._connector_verb = connector_verb
        self._missing_value_fill = missing_value_fill

        # If a `question` was provided and `value_map` was not
        # > infer `value_map` from question (`value_map` is required for `__getitem__`)
        if self._question is not None and self._value_map is None:
            if isinstance(self._question, MultipleChoiceQA):
                self._value_map = self._question.get_value_to_text_map()
            else:
                raise ValueError(
                    f"Cannot infer `ColumnToText` value map from the provided question; "
                    f"Must explicitly provide a `value_map` for column {self.name}.")

        # If `value_map` was provided and `question` was not
        # > infer `question` from value map (if possible)
        elif self._value_map is not None and self._question is None:
            if isinstance(self._value_map, dict):
                self._question = MultipleChoiceQA.create_question_from_value_map(
                    column=self.name,
                    value_map=self._value_map,
                    attribute=self.short_description,
                )

        # Else, warn if both were provided (as they may use inconsistent value maps)
        elif self._value_map is not None and self._question is not None:
            logging.warning(
                f"Got both `value_map` and `question` for column '{self.name}'. "
                f"Please make sure value mappings are consistent.")

        # Else, raise an error if neither were provided
        else:
            raise ValueError(
                f"Must provide either a `value_map` or a `question` for column "
                f"'{self.name}' but neither was provided.")

    @property
    def name(self) -> str:
        return self._name

    @property
    def short_description(self) -> str:
        return self._short_description

    @property
    def question(self) -> QAInterface:
        if self._question is None:
            logging.error(f"No question provided for column '{self.name}'.")
        return self._question

    @property
    def value_map(self) -> Callable:
        """Returns the value map function for this column."""
        if callable(self._value_map):
            return self._value_map
        elif isinstance(self._value_map, dict):
            def _helper_func(value: object) -> str:
                if value not in self._value_map:
                    logging.error(f"Could not find value '{value}' in value map for column '{self.name}'.")
                return self._value_map.get(value, self._missing_value_fill)
            return _helper_func
        else:
            raise ValueError(
                f"Invalid value map type '{type(self._value_map)}' "
                f"for column '{self.name}'.")

    def __getitem__(self, value: object) -> str:
        """Returns the textual representation of the given data value."""
        return self.value_map(value)

    def get_text(self, value: object) -> str:
        return f"The {self.short_description} {self._connector_verb} {self[value]}."



"""Interface for question-answering with LLMs.

- Create different types of questions (direct numeric, multiple-choice).
- Encode questions and decode model outputs.
- Compute risk-estimate from model outputs.
"""
from __future__ import annotations

import dataclasses
import itertools
import logging
from abc import ABC
from dataclasses import dataclass
from typing import Iterator

import torch
from transformers import AutoTokenizer

# Minimum probability density assigned to all valid answers
# > small models will be worse at using valid answers...
ANSWER_PROB_THRESHOLD = 0.1

# Default answer keys for multiple-choice questions
_ALPHABET = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"


@dataclass(frozen=True, eq=True)
class QAInterface(ABC):
    """An interface for a question-answering system."""

    column: str
    text: str
    num_forward_passes: int

    def get_question_prompt(self) -> str:
        """Returns a question and answer key."""
        raise NotImplementedError

    def get_answer_from_model_output(
        self,
        last_token_probs: torch.Tensor,
        tokenizer: AutoTokenizer,
    ) -> float:
        """Decodes the model's output into an answer for the given question.

        Parameters
        ----------
        output : torch.Tensor
            The model's last token probabilities for the question.
        tokenizer : AutoTokenizer
            The tokenizer used to encode the question.

        Returns
        -------
        answer : float
            The answer to the question.
        """
        raise NotImplementedError


@dataclass(frozen=True, eq=True)
class DirectNumericQA(QAInterface):
    """Represents a direct numeric question.

    Notes
    -----
    For example, the prompt could be "
    Q: What is 2 + 2?
    A: "
    With the expected answer being "4".

    If looking for a direct numeric probability, the answer prompt will be
    framed as so: "
    Q: What is the probability, between 0 and 1, of getting heads on a coin flip?
    A: 0."
    So that we can extract a numeric answer with at most 2 forward passes.
    This is done automatically by passing the kwarg `answer_probability=True`.

    Note that some models have multi-digit tokens in their vocabulary, so we
    need to correctly assess which tokens in the vocabulary correspond to valid
    numeric answers.
    """

    num_forward_passes: int = 2     # NOTE: overrides superclass default
    answer_probability: bool = True

    def get_question_prompt(self) -> str:
        question_prompt = f"""Question: {self.text}\n"""
        if self.answer_probability:
            question_prompt += "Answer (between 0 and 1): 0."
        else:
            question_prompt += "Answer: "

        return question_prompt

    def _get_numeric_tokens(self, tokenizer: AutoTokenizer, cache: dict = {}) -> dict[str, int]:
        """Returns the indices of tokens that correspond to numbers.

        At least, this will correspond to one token for each digit 0-9, plus
        some tokenizers have additional tokens for multiple-digit numbers.
        """
        tokenizer_hash = hash(tokenizer)
        if tokenizer_hash not in cache:
            cache[tokenizer_hash] = {
                key: token_id for key, token_id in tokenizer.get_vocab().items()
                if key.isdigit()
            }
        return cache[tokenizer_hash]

    def get_answer_from_model_output(
        self,
        last_token_probs: torch.Tensor,
        tokenizer: AutoTokenizer,
    ) -> float | int:
        """Outputs a numeric answer inferred from the model's output.

        Parameters
        ----------
        last_token_probs : torch.Tensor
            The last token probabilities of the model for the question.
            The first dimension must correspond to the number for forward passes
            as specified by `num_forward_passes`.
        tokenizer : AutoTokenizer
            The tokenizer used to train the model.

        Returns
        -------
        answer : float | int
            The numeric answer to the question.

        Notes
        -----
        Eventually we could run a search algorithm to find the most likely
        answer over multiple forward passes, but for now we'll just take the
        argmax on each forward pass.
        """
        number_to_token_id = self._get_numeric_tokens(tokenizer)

        assert len(last_token_probs) == self.num_forward_passes, (
            f"Expected {self.num_forward_passes} forward passes, got {len(last_token_probs)}."
        )

        answer_text = ""
        # import ipdb; ipdb.set_trace() # TODO: check if this works
        for ltp in last_token_probs:
            # Get the probability of each numeric token
            number_probs = {
                num: ltp[token_id].item()
                for num, token_id in number_to_token_id.items()
            }

            # Get the most likely numeric token
            most_likely_number = max(number_probs, key=number_probs.get)
            answer_text += str(most_likely_number)

            logging.debug(f"Total prob. assigned to numeric tokens: {sum(number_probs.values()):.2%}")

        return int(answer_text) if not self.answer_probability else float(f"0.{answer_text}")


@dataclass(frozen=True, eq=True)
class Choice:
    """Represents a choice in multiple-choice Q&A.

    Attributes
    ----------
    text : str
        The text of the choice. E.g., "25-34 years old".
    data_value : object
        The categorical value corresponding to this choice in the data.
    numeric_value : float, optional
        A meaningful numeric value for the choice. E.g., if the choice is "25-34
        years old", the numeric value could be 30. The choice with the highest
        numeric value can be used as a proxy for the positive class. If not
        provided, will try to use the `choice.value`.
    """
    text: str
    data_value: object
    numeric_value: float = None

    def get_numeric_value(self) -> float:
        """Returns the numeric value of the choice."""
        return self.numeric_value if self.numeric_value is not None else float(self.data_value)


@dataclass(frozen=True, eq=True)    # NOTE: kw_only=True requires Python 3.10
class MultipleChoiceQA(QAInterface):
    """Represents a multiple-choice question and its answer keys."""

    num_forward_passes: int = 1     # NOTE: overrides superclass default
    choices: list[Choice] = dataclasses.field(default_factory=list)
    _answer_keys_source: list[str] = _ALPHABET

    def __post_init__(self):
        if not self.choices:
            raise ValueError("Choices must be provided.")
        if len(self.choices) > len(self._answer_keys_source):
            raise ValueError("Number of choices must be less than or equal to the number of answer keys.")

    @classmethod
    def create_question_from_value_map(
        cls,
        column: str,
        value_map: dict[str, str],
        attribute: str,
        **kwargs,
    ) -> "MultipleChoiceQA":
        """Constructs a question from a value map."""
        choices = [Choice(text, str(value)) for value, text in value_map.items()]

        # Set default question text
        kwargs.setdefault("text", f"What is this person's {attribute}?")

        return cls(
            column=column,
            choices=choices,
            **kwargs,
        )

    @classmethod
    def create_answer_keys_permutations(cls, question: "MultipleChoiceQA") -> Iterator["MultipleChoiceQA"]:
        """Yield questions with all permutations of answer keys.

        Parameters
        ----------
        question : Question
            The template question whose answer keys will be permuted.

        Returns
        -------
        permutations : Iterator[Question]
            A generator of questions with all permutations of answer keys.
        """
        for perm in itertools.permutations(question.choices):
            yield dataclasses.replace(question, choices=perm)

    @property
    def answer_keys(self) -> list[str]:
        return self._answer_keys_source[:len(self.choices)]

    @property
    def key_to_choice(self) -> dict[str, Choice]:
        return dict(zip(self.answer_keys, self.choices))

    @property
    def choice_to_key(self) -> dict[Choice, str]:
        return {choice: key for key, choice in self.key_to_choice.items()}

    def get_value_to_text_map(self) -> dict[object, str]:
        """Returns the map from choice data value to choice textual representation."""
        return {choice.data_value: choice.text for choice in self.choices}

    def get_answer_key_from_value(self, value: object) -> str:
        """Returns the answer key corresponding to the given data value.
        """
        for choice in self.choices:
            if choice.data_value == value:
                return self.choice_to_key[choice]

        logging.error(f"Could not find choice for value: {value}")
        return None

    def get_answer_from_text(self, text: str) -> Choice:
        text = text.strip().upper()
        if text in self.key_to_choice:
            return self.key_to_choice[text]

        logging.error(f"Could not find answer for text: {text}")
        return None

    def get_question_prompt(self) -> str:
        choice_str = "\n".join(
            f"{key}. {choice.text}."
            for key, choice in self.key_to_choice.items()
        )

        return (f"""\
Question: {self.text}
{choice_str}
Answer:""")

    def _decode_model_output_to_choice_distribution(
        self,
        last_token_probs: torch.Tensor,
        tokenizer: AutoTokenizer,
    ) -> float:
        """Decodes the model's output into an answer distribution.

        Parameters
        ----------
        last_token_probs : torch.Tensor
            The model's last token probabilities for the question.
        tokenizer : AutoTokenizer
            The tokenizer used to train the model.

        Returns
        -------
        answers : dict[Choice, float]
            How much probability the model places on each answer choice.

        Notes
        -----
        Answer-key tokens may be prefixed with a space, so we need to check
        both "A" and " A" templates.
        """
        def _get_choice_token_id(choice: Choice, prefix: str = " ") -> int:
            return tokenizer.encode(
                f"{prefix}{self.choice_to_key[choice]}",
                return_tensors="pt",
                add_special_tokens=False,       # don't add BOS or EOS tokens
            ).flatten()[-1]

        prefixes = ["", " "]  # check both "A" and " A" templates

        # Map probabilities to choice values
        answers_per_prefix = {
            prf: {
                choice: last_token_probs[_get_choice_token_id(choice, prefix=prf)].item()
                for choice in self.choices
            }
            for prf in prefixes
        }

        # Choose the prefix with the highest probability density
        best_prefix = max(answers_per_prefix, key=lambda prf: sum(answers_per_prefix[prf].values()))
        answers = answers_per_prefix[best_prefix]

        # Log prefix information in debug mode
        for prefix, choice_probs in answers_per_prefix.items():
            logging.debug(f"prefix='{prefix}' has density {sum(choice_probs.values()):.2%}")

        # Normalize probabilities to sum to 1
        answers_sum_prob = sum(answers.values())

        # Log total probability density assigned to answers
        (logging.warning if answers_sum_prob < ANSWER_PROB_THRESHOLD else logging.debug)(
            f"Answers have {answers_sum_prob:.2%} probability assigned."
        )

        return {
            choice: prob / answers_sum_prob
            for choice, prob in answers.items()
        }

    def get_answer_from_model_output(
        self,
        last_token_probs: torch.Tensor,
        tokenizer: AutoTokenizer,
    ) -> float:
        """Decodes the model's output into an answer for the given question.

        Parameters
        ----------
        output : torch.Tensor
            The model's last token probabilities for the question.
        tokenizer : AutoTokenizer
            The tokenizer used to encode the question.

        Returns
        -------
        answer : float
            The answer to the question.
        """
        answers = self._decode_model_output_to_choice_distribution(
            last_token_probs=last_token_probs,
            tokenizer=tokenizer,
        )

        sorted_choices_by_value = sorted(
            answers.keys(),
            key=lambda choice: choice.get_numeric_value(),
        )

        # If binary question, return probability of positive answer
        # > positive answer always has the highest numeric value
        if len(answers) == 2:
            positive_choice = sorted_choices_by_value[-1]
            return answers[positive_choice]

        # Compute risk estimate by summing weighted choices
        risk_estimate = sum(
            choice.get_numeric_value() * prob
            for choice, prob in answers.items()
        )

        logging.debug(f"Risk estimate: {risk_estimate:.2f}")
        return risk_estimate



"""A collection of instantiated ACS column objects and ACS tasks."""
from __future__ import annotations

from folktexts.qa_interface import DirectNumericQA as _DirectNumericQA
from folktexts.qa_interface import MultipleChoiceQA as _MultipleChoiceQA

from . import acs_columns
from .acs_tasks import _acs_columns_map

# Map of numeric ACS questions
acs_numeric_qa_map: dict[str, object] = {
    question.column: question
    for question in acs_columns.__dict__.values()
    if isinstance(question, _DirectNumericQA)
}

# Map of multiple-choice ACS questions
acs_multiple_choice_qa_map: dict[str, object] = {
    question.column: question
    for question in acs_columns.__dict__.values()
    if isinstance(question, _MultipleChoiceQA)
}

# ... include all multiple-choice questions defined in the column descriptions
acs_multiple_choice_qa_map.update({
    col_to_text.name: col_to_text.question
    for col_to_text in _acs_columns_map.values()
    if col_to_text._question is not None
})
